@@ -20,32 +20,73 @@
  *
  */
 
-#include "gmime-message.h"
-#include "gmime-utils.h"
-#include "internet-address.h"
+
+#ifdef HAVE_CONFIG_H
 #include <config.h>
+#endif
+
 #include <string.h>
 #include <ctype.h>
 #include <locale.h>
 
+#include "gmime-message.h"
+#include "gmime-utils.h"
+#include "gmime-stream-mem.h"
+
+static void g_mime_message_destroy (GMimeObject *object);
+
+static GMimeObject object_template = {
+	0, 0, g_mime_message_destroy
+};
+
+static char *rfc822_headers[] = {
+	"Return-Path",
+	"Received",
+	"Date",
+	"From",
+	"Reply-To",
+	"Subject",
+	"Sender",
+	"To",
+	"Cc",
+	NULL
+};
+
+
 /**
  * g_mime_message_new: Create a new MIME Message object
+ * @pretty_headers: make pretty headers 
  *
- * Returns an empty MIME Message object with no headers nor content
- * set by default.
+ * If @pretty_headers is %TRUE, then the standard rfc822 headers are
+ * initialized so as to put headers in a nice friendly order. This is
+ * strictly a cosmetic thing, so if you are unsure, it is safe to say
+ * no (%FALSE).
+ *
+ * Returns an empty MIME Message object.
  **/
 GMimeMessage *
-g_mime_message_new ()
+g_mime_message_new (gboolean pretty_headers)
 {
 	GMimeMessage *message;
+	GMimeHeader *headers;
+	int i;
 	
 	message = g_new0 (GMimeMessage, 1);
+	g_mime_object_construct (GMIME_OBJECT (message),
+				 &object_template,
+				 GMIME_MESSAGE_TYPE);
 	
 	message->header = g_new0 (GMimeMessageHeader, 1);
 	
 	message->header->recipients = g_hash_table_new (g_str_hash, g_str_equal);
 	
-	message->header->arbitrary_headers = g_ptr_array_new ();
+	message->header->headers = headers = g_mime_header_new ();
+	
+	if (pretty_headers) {
+		/* Populate with the "standard" rfc822 headers so we can have a standard order */
+		for (i = 0; rfc822_headers[i]; i++) 
+			g_mime_header_set (headers, rfc822_headers[i], NULL);
+	}
 	
 	return message;
 }
@@ -53,37 +94,19 @@ g_mime_message_new ()
 static gboolean
 recipients_destroy (gpointer key, gpointer value, gpointer user_data)
 {
-	GList *recipients = value;
+	InternetAddressList *recipients = value;
 	
-	if (recipients) {
-		GList *recipient;
-		
-		recipient = recipients;
-		while (recipient) {
-			internet_address_destroy (recipient->data);
-			recipient = recipient->next;
-		}
-		
-		g_list_free (recipients);
-	}
+	internet_address_list_destroy (recipients);
 	
 	return TRUE;
 }
 
-
-/**
- * g_mime_message_destroy: Destroy the MIME Message.
- * @message: MIME Message to destroy
- *
- * Releases all memory used by the MIME Message and it's child MIME
- * Parts back to the Operating System for reuse.
- **/
-void
-g_mime_message_destroy (GMimeMessage *message)
+static void
+g_mime_message_destroy (GMimeObject *object)
 {
-	int i;
+	GMimeMessage *message = (GMimeMessage *) object;
 	
-	g_return_if_fail (message != NULL);
+	g_return_if_fail (GMIME_IS_MESSAGE (object));
 	
 	g_free (message->header->from);
 	g_free (message->header->reply_to);
@@ -96,28 +119,20 @@ g_mime_message_destroy (GMimeMessage *message)
 	
 	g_free (message->header->message_id);
 	
-	/* destroy arbitrary headers */
-	for (i = 0; i < message->header->arbitrary_headers->len; i++) {
-		GMimeHeader *header = message->header->arbitrary_headers->pdata[i];
-		
-		g_free (header->name);
-		g_free (header->value);
-		g_free (header);
-	}
-	g_ptr_array_free (message->header->arbitrary_headers, TRUE);
+	g_mime_header_destroy (message->header->headers);
 	
 	g_free (message->header);
 	
-	/* destroy all child mime parts */
+	/* unref child mime part */
 	if (message->mime_part)
-		g_mime_part_destroy (message->mime_part);
+		g_mime_object_unref (GMIME_OBJECT (message->mime_part));
 	
 	g_free (message);
 }
 
 
 /**
- * g_mime_message_set_sender: Set the name/address of the sender
+ * g_mime_message_set_sender:
  * @message: MIME Message to change
  * @sender: The name and address of the sender
  *
@@ -125,68 +140,100 @@ g_mime_message_destroy (GMimeMessage *message)
  * (ex: "\"Joe Sixpack\" <joe@sixpack.org>")
  **/
 void
-g_mime_message_set_sender (GMimeMessage *message, const gchar *sender)
+g_mime_message_set_sender (GMimeMessage *message, const char *sender)
 {
-	g_return_if_fail (message != NULL);
+	GMimeMessageHeader *header;
 	
-	if (message->header->from)
-		g_free (message->header->from);
+	g_return_if_fail (GMIME_IS_MESSAGE (message));
 	
-	message->header->from = g_strdup (sender);
+	header = message->header;
+	
+	if (header->from)
+		g_free (header->from);
+	
+	header->from = g_strstrip (g_strdup (sender));
+	g_mime_header_set (header->headers, "From", header->from);
 }
 
 
 /**
- * g_mime_message_get_sender: Get the name/address of the sender
+ * g_mime_message_get_sender:
  * @message: MIME Message
  *
+ * Gets the email address of the sender from @message.
+ *
  * Returns the sender's name and address of the MIME Message.
  **/
-const gchar *
+const char *
 g_mime_message_get_sender (GMimeMessage *message)
 {
-	g_return_val_if_fail (message != NULL, NULL);
+	g_return_val_if_fail (GMIME_IS_MESSAGE (message), NULL);
 	
 	return message->header->from;
 }
 
 
 /**
- * g_mime_message_set_reply_to: Set the Reply-To header of the MIME Message
+ * g_mime_message_set_reply_to:
  * @message: MIME Message to change
  * @reply_to: The Reply-To address
  *
  * Set the sender's Reply-To address on the MIME Message.
  **/
 void
-g_mime_message_set_reply_to (GMimeMessage *message, const gchar *reply_to)
+g_mime_message_set_reply_to (GMimeMessage *message, const char *reply_to)
 {
-	g_return_if_fail (message != NULL);
+	GMimeMessageHeader *header;
 	
-	if (message->header->reply_to)
-		g_free (message->header->reply_to);
+	g_return_if_fail (GMIME_IS_MESSAGE (message));
 	
-	message->header->reply_to = g_strdup (reply_to);
+	header = message->header;
+	
+	if (header->reply_to)
+		g_free (header->reply_to);
+	
+	header->reply_to = g_strstrip (g_strdup (reply_to));
+	g_mime_header_set (header->headers, "Reply-To", header->reply_to);
 }
 
 
 /**
- * g_mime_message_get_reply_to: Get the Reply-To address (if available)
+ * g_mime_message_get_reply_to:
  * @message: MIME Message
  *
+ * Gets the Reply-To address from @message.
+ *
  * Returns the sender's Reply-To address from the MIME Message.
  **/
-const gchar *
+const char *
 g_mime_message_get_reply_to (GMimeMessage *message)
 {
-	g_return_val_if_fail (message != NULL, NULL);
+	g_return_val_if_fail (GMIME_IS_MESSAGE (message), NULL);
 	
 	return message->header->reply_to;
 }
 
 
+static void
+sync_recipient_header (GMimeMessage *message, const char *type)
+{
+	InternetAddressList *recipients;
+	
+	/* sync the specified recipient header */
+	recipients = g_mime_message_get_recipients (message, type);
+	if (recipients) {
+		char *string;
+		
+		string = internet_address_list_to_string (recipients, TRUE);
+		g_mime_header_set (message->header->headers, type, string);
+		g_free (string);
+	} else
+		g_mime_header_set (message->header->headers, type, NULL);
+}
+
+
 /**
- * g_mime_message_add_recipient: Add a recipient
+ * g_mime_message_add_recipient:
  * @message: MIME Message to change
  * @type: Recipient type
  * @name: The recipient's name
@@ -197,23 +244,26 @@ g_mime_message_get_reply_to (GMimeMessage *message)
  * GMIME_RECIPIENT_TYPE_CC and GMIME_RECIPIENT_TYPE_BCC.
  **/
 void
-g_mime_message_add_recipient (GMimeMessage *message, gchar *type, const gchar *name, const gchar *address)
+g_mime_message_add_recipient (GMimeMessage *message, char *type, const char *name, const char *address)
 {
+	InternetAddressList *recipients;
 	InternetAddress *ia;
-	GList *recipients;
 	
-	ia = internet_address_new (name, address);
+	ia = internet_address_new_name (name, address);
 	
 	recipients = g_hash_table_lookup (message->header->recipients, type);
 	g_hash_table_remove (message->header->recipients, type);
 	
-	recipients = g_list_append (recipients, ia);
+	recipients = internet_address_list_append (recipients, ia);
+	internet_address_unref (ia);
+	
 	g_hash_table_insert (message->header->recipients, type, recipients);
+	sync_recipient_header (message, type);
 }
 
 
 /**
- * g_mime_message_add_recipients_from_string: Add a list of recipients
+ * g_mime_message_add_recipients_from_string:
  * @message: MIME Message
  * @type: Recipient type
  * @string: A string of recipient names and addresses.
@@ -225,110 +275,90 @@ g_mime_message_add_recipient (GMimeMessage *message, gchar *type, const gchar *n
  * specified in rfc822.
  **/
 void
-g_mime_message_add_recipients_from_string (GMimeMessage *message, gchar *type, const gchar *string)
+g_mime_message_add_recipients_from_string (GMimeMessage *message, char *type, const char *string)
 {
-	InternetAddress *ia;
-	GList *recipients;
-	gchar *ptr, *eptr, *recipient;
+	InternetAddressList *recipients, *addrlist;
 	
-	g_return_if_fail (message != NULL);
+	g_return_if_fail (GMIME_IS_MESSAGE (message));
 	g_return_if_fail (string != NULL);
 	
 	recipients = g_hash_table_lookup (message->header->recipients, type);
 	g_hash_table_remove (message->header->recipients, type);
 	
-	ptr = (gchar *) string;
-	
-	while (*ptr) {
-		gboolean in_quotes = FALSE;
-		
-		/* skip through leading whitespace */
-		for ( ; *ptr && isspace ((int)*ptr); ptr++);
-		
-		if (*ptr == '"')
-			in_quotes = TRUE;
-		
-		/* find the end of this address */
-		eptr = ptr + 1;
-		while (*eptr) {
-			if (*eptr == '"' && *(eptr - 1) != '\\')
-				in_quotes = !in_quotes;
-			else if (*eptr == ',' && !in_quotes)
-				break;
-			
-			eptr++;
-		}
-		
-		recipient = g_strndup (ptr, (gint) (eptr - ptr));
-		ia = internet_address_new_from_string (recipient);
-		g_free (recipient);
-		recipients = g_list_append (recipients, ia);
-		
-		if (*eptr)
-			ptr = eptr + 1;
-		else
-			break;
-	}
+	addrlist = internet_address_parse_string (string);
+	if (addrlist)
+		recipients = internet_address_list_concat (recipients, addrlist);
 	
 	g_hash_table_insert (message->header->recipients, type, recipients);
+	
+	sync_recipient_header (message, type);
 }
 
 
 /**
- * g_mime_message_get_recipients: Get a list of recipients
+ * g_mime_message_get_recipients:
  * @message: MIME Message
  * @type: Recipient type
  *
+ * Gets a list of recipients of type @type from @message.
+ *
  * Returns a list of recipients of a chosen type from the MIME
  * Message. Available recipient types include:
  * GMIME_RECIPIENT_TYPE_TO, GMIME_RECIPIENT_TYPE_CC and
  * GMIME_RECIPIENT_TYPE_BCC.
  **/
-GList *
-g_mime_message_get_recipients (GMimeMessage *message, const gchar *type)
+InternetAddressList *
+g_mime_message_get_recipients (GMimeMessage *message, const char *type)
 {
-	g_return_val_if_fail (message != NULL, NULL);
+	g_return_val_if_fail (GMIME_IS_MESSAGE (message), NULL);
 	
 	return g_hash_table_lookup (message->header->recipients, type);
 }
 
 
 /**
- * g_mime_message_set_subject: Set the message subject
+ * g_mime_message_set_subject:
  * @message: MIME Message
  * @subject: Subject string
  *
  * Set the Subject field on a MIME Message.
  **/
 void
-g_mime_message_set_subject (GMimeMessage *message, const gchar *subject)
+g_mime_message_set_subject (GMimeMessage *message, const char *subject)
 {
-	g_return_if_fail (message != NULL);
+	GMimeMessageHeader *header;
 	
-	if (message->header->subject)
-		g_free (message->header->subject);
+	g_return_if_fail (GMIME_IS_MESSAGE (message));
 	
-	message->header->subject = g_strdup (subject);
+	header = message->header;
+	
+	if (header->subject)
+		g_free (header->subject);
+	
+	header->subject = g_strstrip (g_strdup (subject));
+	g_mime_header_set (header->headers, "Subject", header->subject);
 }
 
 
 /**
- * g_mime_message_get_subject: Get the message subject
+ * g_mime_message_get_subject:
  * @message: MIME Message
  *
+ * Gets the message's subject.
+ *
  * Returns the Subject field on a MIME Message.
  **/
-const gchar *
+const char *
 g_mime_message_get_subject (GMimeMessage *message)
 {
-	g_return_val_if_fail (message != NULL, NULL);
+	g_return_val_if_fail (GMIME_IS_MESSAGE (message), NULL);
 	
 	return message->header->subject;
 }
 
 
 /**
- * g_mime_message_set_date: Set the message sent-date
+ * g_mime_message_set_date:
  * @message: MIME Message
  * @date: Sent-date (ex: gotten from time (NULL);)
  * @gmt_offset: GMT date offset (in +/- hours)
@@ -338,26 +368,32 @@ g_mime_message_get_subject (GMimeMessage *message)
 void
 g_mime_message_set_date (GMimeMessage *message, time_t date, int gmt_offset)
 {
-	g_return_if_fail (message != NULL);
+	char *date_str;
+	
+	g_return_if_fail (GMIME_IS_MESSAGE (message));
 	
 	message->header->date = date;
 	message->header->gmt_offset = gmt_offset;
+	
+	date_str = g_mime_message_get_date_string (message);
+	g_mime_header_set (message->header->headers, "Date", date_str);
+	g_free (date_str);
 }
 
 
 /**
- * g_mime_message_get_date: Get the message sent-date
+ * g_mime_message_get_date:
  * @message: MIME Message
  * @date: Sent-date
  * @gmt_offset: GMT date offset (in +/- hours)
  * 
  * Stores the date in time_t format in #date and the GMT offset in
- * #gmt_offset.
+ * @gmt_offset.
  **/
 void
 g_mime_message_get_date (GMimeMessage *message, time_t *date, int *gmt_offset)
 {
-	g_return_if_fail (message != NULL);
+	g_return_if_fail (GMIME_IS_MESSAGE (message));
 	
 	*date = message->header->date;
 	*gmt_offset = message->header->gmt_offset;
@@ -365,18 +401,20 @@ g_mime_message_get_date (GMimeMessage *message, time_t *date, int *gmt_offset)
 
 
 /**
- * g_mime_message_get_date_string: Get the message sent-date
+ * g_mime_message_get_date_string:
  * @message: MIME Message
+ *
+ * Gets the message's sent date in string format.
  * 
  * Returns the sent-date of the MIME Message in string format.
  **/
-gchar *
+char *
 g_mime_message_get_date_string (GMimeMessage *message)
 {
-	gchar *date_str;
-	gchar *locale;
+	char *date_str;
+	char *locale;
 	
-	g_return_val_if_fail (message != NULL, NULL);
+	g_return_val_if_fail (GMIME_IS_MESSAGE (message), NULL);
 	
 	locale = g_strdup (setlocale (LC_TIME, ""));
 	setlocale (LC_TIME, "POSIX");
@@ -400,13 +438,19 @@ g_mime_message_get_date_string (GMimeMessage *message)
  * Set the Message-Id on a message.
  **/
 void
-g_mime_message_set_message_id (GMimeMessage *message, const gchar *id)
+g_mime_message_set_message_id (GMimeMessage *message, const char *id)
 {
-	g_return_if_fail (message != NULL);
+	GMimeMessageHeader *header;
 	
-	if (message->header->message_id)
-		g_free (message->header->message_id);
-	message->header->message_id = g_strdup (id);
+	g_return_if_fail (GMIME_IS_MESSAGE (message));
+	
+	header = message->header;
+	
+	if (header->message_id)
+		g_free (header->message_id);
+	
+	header->message_id = g_strstrip (g_strdup (id));
+	g_mime_header_set (header->headers, "Message-Id", header->message_id);
 }
 
 
@@ -414,43 +458,79 @@ g_mime_message_set_message_id (GMimeMessage *message, const gchar *id)
  * g_mime_message_get_message_id: 
  * @message: MIME Message
  *
+ * Gets the Message-Id header of @message.
+ *
  * Returns the Message-Id of a message.
  **/
-const gchar *
+const char *
 g_mime_message_get_message_id (GMimeMessage *message)
 {
-	g_return_val_if_fail (message != NULL, NULL);
+	g_return_val_if_fail (GMIME_IS_MESSAGE (message), NULL);
 	
 	return message->header->message_id;
 }
 
 
 /**
- * g_mime_message_add_arbitrary_header: Add an arbitrary message header
+ * g_mime_message_add_header: Add an arbitrary message header
  * @message: MIME Message
- * @field: rfc822 header field
+ * @header: rfc822 header field
  * @value: the contents of the header field
  *
  * Add an arbitrary message header to the MIME Message such as X-Mailer,
  * X-Priority, or In-Reply-To.
  **/
 void
-g_mime_message_add_arbitrary_header (GMimeMessage *message, const gchar *field, const gchar *value)
+g_mime_message_add_header (GMimeMessage *message, const char *header, const char *value)
 {
-	GMimeHeader *header;
+	g_return_if_fail (GMIME_IS_MESSAGE (message));
+	g_return_if_fail (header != NULL);
 	
-	g_return_if_fail (message != NULL);
-	
-	header = g_new (GMimeHeader, 1);
-	header->name = g_strdup (field);
-	header->value = g_strdup (value);
-	
-	g_ptr_array_add (message->header->arbitrary_headers, header);
+	g_mime_header_add (message->header->headers, header, value);
 }
 
 
 /**
- * g_mime_message_set_mime_part: Set the root-level MIME Part of the message
+ * g_mime_message_set_header: Add an arbitrary message header
+ * @message: MIME Message
+ * @header: rfc822 header field
+ * @value: the contents of the header field
+ *
+ * Set an arbitrary message header to the MIME Message such as X-Mailer,
+ * X-Priority, or In-Reply-To.
+ **/
+void
+g_mime_message_set_header (GMimeMessage *message, const char *header, const char *value)
+{
+	g_return_if_fail (GMIME_IS_MESSAGE (message));
+	g_return_if_fail (header != NULL);
+	
+	g_mime_header_set (message->header->headers, header, value);
+}
+
+
+/**
+ * g_mime_message_get_header:
+ * @message: MIME Message
+ * @header: rfc822 header field
+ *
+ * Gets the value of the requested header @header if it exists, or
+ * %NULL otherwise.
+ *
+ * Returns the value of the requested header (or %NULL if it isn't set)
+ **/
+const char *
+g_mime_message_get_header (GMimeMessage *message, const char *header)
+{
+	g_return_val_if_fail (GMIME_IS_MESSAGE (message), NULL);
+	g_return_val_if_fail (header != NULL, NULL);
+	
+	return g_mime_header_get (message->header->headers, header);
+}
+
+
+/**
+ * g_mime_message_set_mime_part:
  * @message: MIME Message
  * @mime_part: The root-level MIME Part
  *
@@ -459,174 +539,66 @@ g_mime_message_add_arbitrary_header (GMimeMessage *message, const gchar *field,
 void
 g_mime_message_set_mime_part (GMimeMessage *message, GMimePart *mime_part)
 {
-	g_return_if_fail (message != NULL);
+	g_return_if_fail (GMIME_IS_MESSAGE (message));
+	g_return_if_fail (GMIME_IS_PART (mime_part));
+	
+	g_mime_object_ref (GMIME_OBJECT (mime_part));
 	
 	if (message->mime_part)
-		g_mime_part_destroy (message->mime_part);
+		g_mime_object_unref (GMIME_OBJECT (message->mime_part));
 	
 	message->mime_part = mime_part;
 }
 
 
-static gchar *
-create_header (GMimeMessage *message)
+/**
+ * g_mime_message_write_to_stream:
+ * @message: MIME Message
+ * @stream: output stream
+ *
+ * Write the contents of the MIME Message to @string.
+ **/
+void
+g_mime_message_write_to_stream (GMimeMessage *message, GMimeStream *stream)
 {
-	/* Make sure to not append "\n\n" to the end
-	 * so we can add MIME headers later */
-	GString *string;
-	gchar *str, *buf, *date, *subject;
-	GList *recipients;
+	g_return_if_fail (GMIME_IS_MESSAGE (message));
+	g_return_if_fail (stream != NULL);
 	
-	string = g_string_new ("");
+	g_mime_header_write_to_stream (message->header->headers, stream);
 	
-	/* write out the arbitrary headers first (as they may contain
-	 * "Received:" headers which really should come first) */
-	if (message->header->arbitrary_headers->len) {
-		gint i;
-		
-		for (i = 0; i < message->header->arbitrary_headers->len; i++) {
-			const GMimeHeader *header;
-			gchar *encoded_value;
-			
-			header = message->header->arbitrary_headers->pdata[i];
-			encoded_value = g_mime_utils_8bit_header_encode (header->value);
-			
-			buf = g_mime_utils_header_printf ("%s: %s\n", header->name, encoded_value);
-			g_free (encoded_value);
-			
-			g_string_append (string, buf);
-			g_free (buf);
-		}
-	}
-	
-	/* create the standard headers */
-	if (!message->header->date)
-		g_mime_message_set_date (message, time (NULL), 0);
-	date = g_mime_message_get_date_string (message);
-	buf = g_mime_utils_header_printf ("Date: %s\n", date);
-	g_string_append (string, buf);
-	g_free (date);
-	g_free (buf);
-	
-	buf = g_mime_utils_header_printf ("From: %s\n", message->header->from ? message->header->from : "");
-	g_string_append (string, buf);
-	g_free (buf);
-	
-	if (message->header->reply_to) {
-		buf = g_mime_utils_header_printf ("Reply-To: %s\n", message->header->reply_to);
-		g_string_append (string, buf);
-		g_free (buf);
-	}
-	
-	recipients = g_mime_message_get_recipients (message, GMIME_RECIPIENT_TYPE_TO);
-	if (recipients) {
-		GString *recip;
-		
-		recip = g_string_new ("To: ");
-		while (TRUE) {
-			InternetAddress *ia;
-			gchar *address;
-			
-			ia = recipients->data;
-			address = internet_address_to_string (ia, TRUE);
-			g_string_append (recip, address);
-			g_free (address);
-			
-			recipients = recipients->next;
-			if (recipients)
-				g_string_append (recip, ", ");
-			else
-				break;
-		}
-		g_string_append (recip, "\n");
-		buf = g_mime_utils_header_fold (recip->str);
-		g_string_free (recip, TRUE);
-		
-		g_string_append (string, buf);
-		g_free (buf);
-	}
-	
-	recipients = g_mime_message_get_recipients (message, GMIME_RECIPIENT_TYPE_CC);
-	if (recipients) {
-		GString *recip;
-		
-		recip = g_string_new ("Cc: ");
-		while (TRUE) {
-			InternetAddress *ia;
-			gchar *address;
-			
-			ia = recipients->data;
-			address = internet_address_to_string (ia, TRUE);
-			g_string_append (recip, address);
-			g_free (address);
-			
-			recipients = recipients->next;
-			if (recipients)
-				g_string_append (recip, ", ");
-			else
-				break;
-		}
-		g_string_append (recip, "\n");
-		buf = g_mime_utils_header_fold (recip->str);
-		g_string_free (recip, TRUE);
-		
-		g_string_append (string, buf);
-		g_free (buf);
-	}
-	
-	subject = g_mime_utils_8bit_header_encode (message->header->subject);
-	buf = g_mime_utils_header_printf ("Subject: %s\n", subject ? subject : "");
-	g_string_append (string, buf);
-	g_free (subject);
-	g_free (buf);
-	
-	if (message->header->message_id) {
-		gchar *message_id;
-		
-		message_id = g_mime_utils_8bit_header_encode (message->header->message_id);
-		buf = g_mime_utils_header_printf ("Message-Id: %s\n", message_id ? message_id : "");
-		g_string_append (string, buf);
-		g_free (message_id);
-		g_free (buf);
-	}
-	
-	str = string->str;
-	g_string_free (string, FALSE);
-	
-	return str;
+	if (message->mime_part) {
+		g_mime_stream_write_string (stream, "MIME-Version: 1.0\n");
+		g_mime_part_write_to_stream (message->mime_part, stream);
+	} else
+		g_mime_stream_write (stream, "\n", 1);
 }
 
 
 /**
- * g_mime_message_to_string: Write the MIME Message to a string
+ * g_mime_message_to_string:
  * @message: MIME Message
  *
+ * Allocates a string buffer containing the mime message @message.
+ *
  * Returns an allocated string containing the MIME Message.
  **/
-gchar *
+char *
 g_mime_message_to_string (GMimeMessage *message)
 {
-	GString *string;
-	gchar *str, *header, *body;
+	GMimeStream *stream;
+	GByteArray *array;
+	char *str;
 	
-	g_return_val_if_fail (message != NULL, NULL);
+	g_return_val_if_fail (GMIME_IS_MESSAGE (message), NULL);
 	
-	header = create_header (message);
-	if (!header)
-		return NULL;
-	
-	string = g_string_new (header);
-	g_free (header);
-	
-	body = g_mime_part_to_string (message->mime_part, TRUE);
-	if (body)
-		g_string_append (string, body);
-	else
-		g_string_append (string, "\n");
-	g_free (body);
-	
-	str = string->str;
-	g_string_free (string, FALSE);
+	array = g_byte_array_new ();
+	stream = g_mime_stream_mem_new ();
+	g_mime_stream_mem_set_byte_array (GMIME_STREAM_MEM (stream), array);
+	g_mime_message_write_to_stream (message, stream);
+	g_mime_stream_unref (stream);
+	g_byte_array_append (array, "", 1);
+	str = array->data;
+	g_byte_array_free (array, FALSE);
 	
 	return str;
 }
@@ -661,14 +633,14 @@ g_mime_message_to_string (GMimeMessage *message)
  * message body. Note that this is an assumption and is thus not
  * guarenteed to always be correct.
  **/
-static gchar *
+static char *
 multipart_get_body (GMimePart *multipart, gboolean want_plain, gboolean *is_html)
 {
 	GMimePart *first = NULL;
-	const gchar *content;
-	gchar *body = NULL;
+	const char *content;
+	char *body = NULL;
 	GList *child;
-	guint len;
+	size_t len;
 	
 	child = multipart->children;
 	while (child) {
@@ -714,26 +686,33 @@ multipart_get_body (GMimePart *multipart, gboolean want_plain, gboolean *is_html
 	return body;
 }
 
+
 /**
- * g_mime_message_get_body: Return the body of the message
+ * g_mime_message_get_body:
  * @message: MIME Message
  * @want_plain: request text/plain
  * @is_html: body returned is in html format
  *
+ * Attempts to get the body of the message in the preferred format
+ * specified by @want_plain.
+ *
  * Returns the prefered form of the message body. Sets the value of
- * #is_html to TRUE if the part returned is in HTML format, otherwise
+ * @is_html to TRUE if the part returned is in HTML format, otherwise
  * FALSE.
  * Note: This function is NOT guarenteed to always work as it
  * makes some assumptions that are not necessarily true. It is
  * recommended that you traverse the MIME structure yourself.
  **/
-gchar *
+char *
 g_mime_message_get_body (const GMimeMessage *message, gboolean want_plain, gboolean *is_html)
 {
 	const GMimeContentType *type;
-	const gchar *content;
-	gchar *body = NULL;
-	guint len = 0;
+	const char *content;
+	char *body = NULL;
+	size_t len = 0;
+
+	g_return_val_if_fail (message!=NULL, NULL);
+	g_return_val_if_fail (is_html!=NULL, NULL);
 	
 	type = g_mime_part_get_content_type (message->mime_part);
 	if (g_mime_content_type_is_type (type, "text", "*")) {
@@ -755,15 +734,19 @@ g_mime_message_get_body (const GMimeMessage *message, gboolean want_plain, gbool
 
 
 /**
- * g_mime_message_get_headers: Return the raw message headers
+ * g_mime_message_get_headers:
  * @message: MIME Message
  *
+ * Allocates a string buffer containing the raw message headers.
+ *
  * Returns an allocated string containing the raw message headers.
  **/
-gchar *
+char *
 g_mime_message_get_headers (GMimeMessage *message)
 {
-	return create_header (message);
+	g_return_val_if_fail (GMIME_IS_MESSAGE (message), NULL);
+	
+	return g_mime_header_to_string (message->header->headers);
 }
 
 
@@ -773,12 +756,12 @@ g_mime_message_get_headers (GMimeMessage *message)
  * @callback: function to call on each of the mime parts contained by the mime message
  * @data: extra data to pass to the callback
  *
- * Calls #callback on each of the mime parts in the mime message.
+ * Calls @callback on each of the mime parts in the mime message.
  **/
 void
 g_mime_message_foreach_part (GMimeMessage *message, GMimePartFunc callback, gpointer data)
 {
-	g_return_if_fail (message != NULL);
+	g_return_if_fail (GMIME_IS_MESSAGE (message));
 	g_return_if_fail (callback != NULL);
 	
 	g_mime_part_foreach (message->mime_part, callback, data);
