@@ -81,7 +81,8 @@ void articlelist_read_message_id           (const gchar * message_id);
 **/
 
 void articlelist_set_group_nolock (Group*);
-void articlelist_refresh_nolock (void);
+void articlelist_refresh (void);
+void articlelist_read_selected (void);
 
 /**
 ***  Perform Operations on Articles
