@@ -20,6 +20,7 @@
 
 #include <config.h>
 
+#include <ctype.h>
 #include <stdlib.h>
 #include <string.h>
 
@@ -32,6 +33,7 @@
 #include <pan/base/group.h>
 #include <pan/base/pan-glib-extensions.h>
 #include <pan/base/pan-i18n.h>
+#include <pan/base/run.h>
 #include <pan/base/server.h>
 #include <pan/base/group.h>
 #include <pan/base/util-mime.h>
@@ -81,6 +83,7 @@ article_constructor (Article * article, Group * group)
 
 	/* init this class */
 	article->group = group;
+	article->is_new = FALSE;
 	article->self_passes_filter = FALSE;
 	article->tree_passes_filter = FALSE;
 	article->part = (gint16)0;
@@ -171,6 +174,7 @@ static void
 articles_set_dirty (Article ** articles, gint qty)
 {
 	Group * group;
+	debug_enter ("articles_set_dirty");
 
 	/* sanity check */
 	g_return_if_fail (qty >= 1);
@@ -180,6 +184,8 @@ articles_set_dirty (Article ** articles, gint qty)
 	group = articles[0]->group;
 	group_set_articles_dirty (group);
 	fire_articles_changed (group, articles, qty, ARTICLE_CHANGED_DIRTY);
+
+	debug_exit ("articles_set_dirty");
 }
 
 
@@ -188,9 +194,55 @@ article_is_new (const Article * a)
 {
 	g_return_val_if_fail (article_is_valid(a), FALSE);
 
-	return a->number > a->group->article_high_old;
+	return a->is_new && !article_is_read(a);
 }
 
+/***
+****  NEW / OLD
+***/
+
+static void
+articles_set_new_impl (Article ** articles, int article_qty, gboolean is_new, gboolean fire)
+{
+	int i;
+	GPtrArray * changed;
+	debug_enter ("articles_set_new_impl");
+
+	/* sanity check */
+	g_return_if_fail (article_qty > 0);
+	g_return_if_fail (articles_are_valid ((const Article **)articles, article_qty));
+
+	/* mark each article's newness field */
+	changed = g_ptr_array_new ();
+	for (i=0; i<article_qty; ++i) {
+		Article * a = articles[i];
+		if (!!is_new != !!a->is_new) {
+			a->is_new = is_new;
+			g_ptr_array_add (changed, a);
+		}
+	}
+
+	/* update parent counts of changed articles */
+	for (i=0; i<changed->len; ++i) {
+		Article * a = ARTICLE(g_ptr_array_index(changed, i));
+		for (a=a->parent; a!=NULL; a=a->parent)
+			a->new_children += is_new ? 1 : -1;
+	}
+
+	/* fire notification of changed articles */
+	if (fire && changed->len!=0)
+		fire_articles_changed (articles[0]->group, (Article**)changed->pdata, changed->len, ARTICLE_CHANGED_NEW);
+
+	/* cleanup */
+	g_ptr_array_free (changed, TRUE);
+	debug_exit ("articles_set_new_impl");
+}
+
+void
+articles_set_new (Article ** articles, int article_qty, gboolean is_new)
+{
+	articles_set_new_impl (articles, article_qty, is_new, TRUE);
+}
 
 /***
 ****
@@ -217,16 +269,26 @@ articles_set_read_numbers_ghfunc (gpointer key, gpointer value, gpointer user_da
 	gboolean read = user_data!=NULL;
 	Newsrc * newsrc = group_get_newsrc (g);
 
-	/* for each number, mark it read */
-	for (i=changed_qty=0; i<numbers->len; ++i) {
+	/**
+	***  Mark each number as read/unread
+	**/
+
+	for (i=changed_qty=0; i<numbers->len; ++i)
+	{
 		const gulong number = g_array_index (numbers, gulong, i);
+
 		if (read != newsrc_mark_article (newsrc, number, read))
 			++changed_qty;
 	}
 
-	/* if changed, alter the group's read qty */
+	/**
+	***  If any changed, update the group's count
+	**/
+
 	if (changed_qty != 0)
+	{
 		group_inc_article_read_qty (g, read ? changed_qty : -changed_qty);
+	}
 
 	g_array_free (numbers, TRUE);
 }
@@ -234,26 +296,53 @@ articles_set_read_numbers_ghfunc (gpointer key, gpointer value, gpointer user_da
 static void
 articles_set_read_articles_ghfunc (gpointer key, gpointer value, gpointer user_data)
 {
-	gint i;
-	GPtrArray * changed = g_ptr_array_new ();
+	guint i;
 	Group * g = GROUP(key);
 	GArray * articles = (GArray*) value;
 	gboolean read = user_data!=NULL;
 	Newsrc * newsrc = group_get_newsrc (g);
+	GPtrArray * changed = g_ptr_array_new ();
 
-	/* for each article, mark it read */
-	for (i=0; i<articles->len; ++i) {
+	/**
+	***  Walk through the articles & make a list of those to be changed
+	**/
+
+	for (i=0; i!=articles->len; ++i)
+	{
 		Article * a = ARTICLE(g_array_index(articles,Article*,i));
-		if (read != newsrc_mark_article (newsrc, a->number, read))
+
+		if (!!read != !!newsrc_mark_article (newsrc, a->number, read))
 			g_ptr_array_add (changed, a);
 	}
 
-	/* if changed, alter the group's read qty & fire an event */
-	if (changed->len != 0) {
+	/**
+	***  If anything to be changed:
+	***  (1) update parent counts
+	***  (2) update group's counts
+	***  (3) fire an event
+	**/
+
+	if (changed->len != 0)
+	{
+		/* update parent counts */
+		for (i=0; i!=changed->len; ++i)
+		{
+			Article * a = ARTICLE(g_ptr_array_index(changed,i));
+			for (a=a->parent; a!=NULL; a=a->parent)
+				a->unread_children += read ? -1 : 1;
+		}
+		if (read)
+			articles_set_new_impl ((Article**)changed->pdata, changed->len, FALSE, FALSE);
+
+		/* update the group's count */
 		group_inc_article_read_qty (g, read ? changed->len : -changed->len);
+
+		/* fire an event */
 		fire_articles_changed (g, (Article**)changed->pdata, changed->len, ARTICLE_CHANGED_READ);
 	}
 
+
+	/* cleanup */
 	g_array_free (articles, TRUE);
 }
 
@@ -302,29 +391,35 @@ articles_set_read (Article ** articles, int article_qty, gboolean read)
 	g_return_if_fail (article_qty > 0);
 	g_return_if_fail (articles_are_valid ((const Article **)articles, article_qty));
 
-	/* get a big list, grouped by Group*,  of all the articles that we're marking read */
+	/* get a big hash, keyed by Group*,  of all the articles that we're
+	   marking read.  We have two hashes -- one for the articles passed
+	   in, and one for article numbers of the xposts.  */
 	group_to_articles = g_hash_table_new (g_direct_hash, g_direct_equal);
 	for (i=0; i<article_qty; ++i)
 	{
-		GSList * l;
 		GArray * arr;
 		Article * a = articles[i];
 
-		/* add this article to the group's array */
+		/* first add this article to the group's array */
 	       	arr = get_group_array_from_hash (group_to_articles, a->group, sizeof(Article*));
 		g_array_append_val (arr, a);
-		/* add the xrefs to their group's arrays */
+
+		/* secondly, add the xrefs to their group's arrays */
 		article_xref_foreach (a, mark_read_xreffunc, &group_to_numbers,
 		                      SERVER_GROUPS_SUBSCRIBED, TRUE);
 
-		/* if it's a multipart, process the child parts too */
-		if (a->part>0 && a->parts>1 && a->threads!=NULL)
+		/* lastly, if it's a multipart, process the child parts too */
+		if (a->part==1 && a->parts>1 && a->threads!=NULL)
 		{
-			for (l=a->threads; l!=NULL; l=l->next) {
+			GSList * l;
+			for (l=a->threads; l!=NULL; l=l->next)
+			{
 				Article * child = ARTICLE(l->data);
-				if (child->part > 0) {
+				if (child->part>a->part && child->parts==a->parts)
+				{
 					/* add this article to the group's array */
 					g_array_append_val (arr, child);
+
 					/* add the xrefs to their group's arrays */
 					article_xref_foreach (child,
 						mark_read_xreffunc,
@@ -336,12 +431,17 @@ articles_set_read (Article ** articles, int article_qty, gboolean read)
 	}
 
 	/* mark the passed-in articles as read/unread */
-	g_hash_table_foreach (group_to_articles, articles_set_read_articles_ghfunc, GINT_TO_POINTER(read));
+	g_hash_table_foreach (group_to_articles,
+	                      articles_set_read_articles_ghfunc,
+	                      GINT_TO_POINTER(read));
 	g_hash_table_destroy (group_to_articles);
 
+
 	/* mark their crosspost counterparts, if any, as read/unread */
 	if (group_to_numbers != NULL) {
-		g_hash_table_foreach (group_to_numbers, articles_set_read_numbers_ghfunc, GINT_TO_POINTER(read));
+		g_hash_table_foreach (group_to_numbers,
+		                      articles_set_read_numbers_ghfunc,
+		                      GINT_TO_POINTER(read));
 		g_hash_table_destroy (group_to_numbers);
 	}
 
@@ -359,16 +459,26 @@ articles_add_flag (Article  ** articles,
                    gushort     flag)
 {
 	gint i;
+	gint changed_qty;
+	Article ** changed;
 
 	/* sanity clause */
 	g_return_if_fail (qty >= 1);
 	g_return_if_fail (articles_are_valid ((const Article **)articles, qty));
 
 	/* mark 'em */
-	for (i=0; i<qty; ++i)
-		articles[i]->state |= flag;
+	changed = g_malloc (sizeof(Article*) * qty);
+	for (changed_qty=i=0; i<qty; ++i) {
+		if (!(articles[i]->state & flag)) {
+			articles[i]->state |= flag;
+			changed[changed_qty++] = articles[i];
+		}
+	}
 
-	articles_set_dirty (articles, qty);
+	if (changed_qty > 0)
+		articles_set_dirty (changed, changed_qty);
+
+	g_free (changed);
 }
 
 
@@ -378,16 +488,26 @@ articles_remove_flag (Article         ** articles,
                       gushort            flag)
 {
 	gint i;
+	gint changed_qty;
+	Article ** changed;
 
 	/* sanity clause */
 	g_return_if_fail (qty >= 1);
 	g_return_if_fail (articles_are_valid ((const Article **)articles, qty));
 
 	/* mark 'em */
-	for (i=0; i<qty; ++i)
-		articles[i]->state &= ~flag;
+       	changed = g_malloc (sizeof(Article*) * qty);
+	for (changed_qty=i=0; i<qty; ++i) {
+		if (articles[i]->state & flag) {
+			articles[i]->state &= ~flag;
+			changed[changed_qty++] = articles[i];
+		}
+	}
 
-	articles_set_dirty (articles, qty);
+	if (changed_qty > 0)
+		articles_set_dirty (changed, changed_qty);
+
+	g_free (changed);
 }
 
 
@@ -412,7 +532,7 @@ article_get_thread_message_id (const Article* article)
 	while (article->parent != NULL)
 		article = article->parent;
 
-	refs = article_get_header (article, HEADER_REFERENCES);
+	refs = article->references;
 	if (is_nonempty_string(refs))
 	{
 		pch = get_next_token_str (refs, ' ', NULL);
@@ -443,6 +563,10 @@ article_get_thread_impl (Article * article,
 {
 	Article* top = article;
 
+	/* sanity clause */
+	g_return_if_fail (article_is_valid(article));
+	g_return_if_fail (setme!=NULL);
+
 	if (top != NULL)
 	{
 		if (!children_only)
@@ -469,8 +593,8 @@ article_get_entire_thread (Article* article, GPtrArray* setme)
 void
 article_get_references (Article* article, GPtrArray* setme)
 {
+	g_return_if_fail (article_is_valid(article));
 	g_return_if_fail (setme!=NULL);
-	g_return_if_fail (article!=NULL);
 
 	pan_g_ptr_array_reserve (setme, 128);
 	while (article != NULL) {
@@ -484,12 +608,20 @@ article_get_unique_threads (const GPtrArray    * articles,
                             ThreadGet            thread_get)
 {
 	int i;
-	GPtrArray * thread = g_ptr_array_new ();
-	GPtrArray * retval = g_ptr_array_new ();
-	GHashTable * all = g_hash_table_new (g_str_hash, g_str_equal);
+	GPtrArray * thread;
+	GPtrArray * retval;
+	GHashTable * all;
 	debug_enter ("articlelist_get_unique_threads");
 
+	/* sanity clause */
+	retval = g_ptr_array_new ();
+	g_return_val_if_fail (articles_are_valid ((const Article **)articles->pdata, articles->len), retval);
+
 	/* get the thread for each article passed in */
+	thread = g_ptr_array_new ();
+	pan_g_ptr_array_reserve (retval, 128);
+	pan_g_ptr_array_reserve (thread, 128);
+	all = g_hash_table_new (g_str_hash, g_str_equal);
 	for (i=0; i<articles->len; ++i)
 	{
 		int j;
@@ -565,7 +697,7 @@ article_remove_header (Article * a, const gchar * key)
 }
 
 const gchar*
-article_get_header (const Article * a, const gchar * key)
+article_get_extra_header (const Article * a, const gchar * key)
 {
 	const gchar * retval = NULL;
 
@@ -573,17 +705,7 @@ article_get_header (const Article * a, const gchar * key)
 	g_return_val_if_fail (article_is_valid(a), NULL);
 	g_return_val_if_fail (is_nonempty_string(key), NULL);
 
-	if (key == HEADER_MESSAGE_ID) retval = a->message_id;
-	else if (key == HEADER_SUBJECT) retval = a->subject;
-	else if (key == HEADER_REFERENCES) retval = a->references;
-	else if (key == HEADER_XREF) retval = a->xref;
-	else if (key == HEADER_FROM) pan_warn_if_reached ();
-	else if (!strcmp(key, HEADER_MESSAGE_ID)) retval = a->message_id;
-	else if (!strcmp (key, HEADER_SUBJECT)) retval = a->subject;
-	else if (!strcmp (key, HEADER_REFERENCES)) retval = a->references;
-	else if (!strcmp (key, HEADER_XREF)) retval = a->xref;
-	else if (!strcmp (key, HEADER_FROM)) pan_warn_if_reached ();
-	else if (a->headers!=NULL)
+	if (a->headers!=NULL)
 		retval = (const gchar*) g_hash_table_lookup (a->headers, key);
 
 	return retval;
@@ -628,44 +750,38 @@ article_xref_foreach (const Article       * a,
 		      ServerGroupsType      set,
                       gboolean              skip_group_a)
 {
-	gchar * str;
-	Server * server;
-	const gchar * xref;
-	const gchar * march;
-
 	/* sanity clause */
 	g_return_if_fail (article_is_valid(a));
-	g_return_if_fail (a->group!=NULL);
-	g_return_if_fail (a->group->server!=NULL);
 
 	/* get the xref header */
-	xref = article_get_header (a, HEADER_XREF);
-	if (!is_nonempty_string(xref))
-		return;
-
-	/* skip the servername; we've got the server already */
-	skip_next_token (xref, ' ', &march);
-
-	/* walk through the xrefs, of format "group1:number group2:number" */
-	server = a->group->server;
-	while ((str = get_next_token_str (march, ' ', &march)))
+	if (is_nonempty_string(a->xref))
 	{
-		gchar * delimit = strchr (str, ':');
+		Run run;
+		const gchar * march;
+		Server * const server = a->group->server;
 
-		if (delimit != NULL)
+		/* skip the servername; we've got the server already */
+		skip_next_token (a->xref, ' ', &march);
+
+		/* walk through the xrefs, of format "group1:number group2:number" */
+		while (get_next_token_run (march, ' ', &march, &run.str, &run.len))
 		{
-			const gulong number = strtoul (delimit+1, NULL, 10);
-			Group * g = NULL;
-			*delimit = '\0';
+			const gchar * delimit = run_strchr (&run, ':');
+			if (delimit != NULL)
+			{
+				const gulong number = strtoul (delimit+1, NULL, 10);
+				Group * g = NULL;
+				gchar * group_name;
 
-			if (server_has_group_in_type (server, str, set))
-				g = server_get_named_group (server, str);
+				pan_strndup_alloca (group_name, run.str, delimit-run.str);
 
-			if (g!=NULL && (a->group!=g || !skip_group_a))
-				(*func)(g, number, user_data);
+				if (server_has_group_in_type (server, group_name, set))
+					g = server_get_named_group (server, group_name);
+
+				if (g!=NULL && (a->group!=g || !skip_group_a))
+					(*func)(g, number, user_data);
+			}
 		}
-
-		g_free (str);
 	}
 }
 
@@ -763,7 +879,7 @@ article_has_attachment (const Article *a)
 {
 	g_return_val_if_fail (article_is_valid(a), FALSE);
 
-	return article_get_header(a,PAN_ATTACH_FILE) != NULL;
+	return article_get_extra_header(a,PAN_ATTACH_FILE) != NULL;
 }
 
 /***
@@ -779,53 +895,41 @@ article_has_body (const Article * a)
 
 	g_return_val_if_fail (article_is_valid(a), FALSE);
 
-	/* maybe the article's in a folder so the body
-	 * is hidden as a header... */
-	if (!retval) {
-		const gchar * body = article_get_header (a, PAN_BODY);
-		if (body != NULL)
-			retval = TRUE;
-	}
-
 	/* try to get it from acache... */
 	if (!retval && is_nonempty_string(a->message_id))
 		retval = acache_has_message (a->message_id);
 
+	/* maybe the article's in a folder so the body
+	 * is hidden as a header... */
+	if (!retval)
+		if (article_get_extra_header (a, PAN_BODY) != NULL)
+			retval = TRUE;
+
 	return retval;
 }
 
 gchar*
-article_get_author_str (const Article * a)
+article_get_author_str (const Article * a, gchar * buf, gint bufsize)
 {
-        gchar * retval;
+	const gboolean have_addr = a!=NULL && is_nonempty_string (a->author_addr);
+	const gboolean have_name = a!=NULL && is_nonempty_string (a->author_real);
 
-	if (a==NULL)
-	{
-		retval = g_strdup ("");
+	*buf = '\0';
+
+	if (have_addr && have_name) {
+		g_snprintf (buf, bufsize, "\"%s\" <%s>", a->author_real, a->author_addr);
 	}
-	else if (is_nonempty_string(a->author_addr))
-	{
-		if (is_nonempty_string(a->author_real))
-		{
-			retval = g_strdup_printf ("\"%s\" <%s>", a->author_real, a->author_addr);
-		}
-		else
-		{
-			retval = g_strdup (a->author_addr);
-		}
+	else if (have_addr) {
+		strncpy (buf, a->author_addr, bufsize);
+		buf[bufsize-1] = '\0';
 	}
-	else if (is_nonempty_string(a->author_real))
-	{
-		retval = g_strdup (a->author_real);
-	}
-	else
-	{
-		retval = g_strdup ("");
+	else if (have_name) {
+		strncpy (buf, a->author_real, bufsize);
+		buf[bufsize-1] = '\0';
 	}
 
-	g_strstrip (retval);
-
-	return retval;
+	g_strstrip (buf);
+	return buf;
 }
 
 void
@@ -863,14 +967,16 @@ article_get_message (const Article * a)
 	gchar * retval = NULL;
 
 	/* sanity clause */
-	g_return_val_if_fail (article_is_valid(a), NULL);
+	g_return_val_if_fail (article_is_valid(a), g_strdup(""));
 
-	if (retval==NULL && a!=NULL && acache_has_message (a->message_id))
+	if (retval==NULL)
 	{
-		retval = acache_get_message (a->message_id);
+		const gchar * message_id = article_get_message_id (a);
+		if (acache_has_message (message_id))
+			retval = acache_get_message (message_id);
 	}
 
-	if (retval==NULL && a!=NULL)
+	if (retval==NULL)
 	{
 		GString * s = g_string_new (NULL);
 		gchar * pch;
@@ -900,32 +1006,39 @@ article_get_headers (const Article * a)
 	g_return_val_if_fail (article_is_valid(a), g_strdup(""));
 
 	/* do we have it in acache? */
-	if (retval==NULL && a!=NULL && acache_has_message (a->message_id))
+	if (retval==NULL)
 	{
-		gchar * s = acache_get_message (a->message_id);
-		gchar * pch = pan_strstr (s, "\n\n");
-		if (pch != NULL)
-			retval = g_strndup (s, (gint)(pch-s));
-		g_free (s);
+		const gchar * message_id = article_get_message_id (a);
+		if (acache_has_message (message_id))
+		{
+			gchar * s = acache_get_message (message_id);
+			gchar * pch = pan_strstr (s, "\n\n");
+			if (pch != NULL) {
+				*pch = '\0';
+				retval = s;
+			}
+		}
 	}
 
 	/* can we build it from a? */
-	if (retval==NULL && a!=NULL)
+	if (retval==NULL)
 	{
 		gchar * tmp;
 		const gchar * c_tmp;
 		GString * s = g_string_new (NULL);
 
 		/* subject */
-		c_tmp = a->subject;
+		c_tmp = article_get_subject (a);
 		if (is_nonempty_string (c_tmp))
 			g_string_sprintfa (s, "%s: %s\n", HEADER_SUBJECT, c_tmp);
 
 		/* author */
-	       	tmp = article_get_author_str (a);
-		if (is_nonempty_string (tmp))
-			g_string_sprintfa (s, "%s: %s\n", HEADER_FROM, tmp);
-		g_free (tmp);
+		if (1) {
+			gchar author[512] = { '\0' };
+	       		article_get_author_str (a, author, sizeof(author));
+			if (is_nonempty_string (author))
+				g_string_sprintfa (s, "%s: %s\n", HEADER_FROM, author);
+		}
 
 		/* date */
 		if (a->date != 0) {
@@ -935,7 +1048,7 @@ article_get_headers (const Article * a)
 		}
 
 		/* message-id */
-		c_tmp = a->message_id;
+		c_tmp = article_get_message_id (a);
 		if (is_nonempty_string (c_tmp))
 			g_string_sprintfa (s, "%s: %s\n", HEADER_MESSAGE_ID, c_tmp);
 
@@ -969,38 +1082,111 @@ article_get_body (const Article * a)
 {
 	gchar * retval = NULL;
 
-	g_return_val_if_fail (article_is_valid(a), NULL);
+	g_return_val_if_fail (article_is_valid(a), g_strdup(""));
 
 	/* see if we've got the body hidden in a header... */
-	if (retval==NULL && a!=NULL)
+	if (retval==NULL)
 	{
-		const gchar * body = article_get_header (a, PAN_BODY);
+		const gchar * body = article_get_extra_header (a, PAN_BODY);
 		if (body != NULL)
 			retval = g_strdup (body);
 	}
 
 	/* see if we've got it in acache... */
-	if (retval==NULL && a!=NULL && acache_has_message(a->message_id))
+	if (retval==NULL)
 	{
-		gchar * s = acache_get_message (a->message_id);
-		gboolean is_html = FALSE;
-		GMimeMessage * m = pan_g_mime_parser_construct_message (s);
-		retval = g_mime_message_get_body (m, TRUE, &is_html);
-		g_mime_message_destroy (m);
-		g_free (s);
+		const gchar * message_id = article_get_message_id (a);
+		if (acache_has_message (message_id))
+		{
+			gchar * s = acache_get_message (message_id);
+			gboolean is_html = FALSE;
+			GMimeMessage * m = pan_g_mime_parser_construct_message (s);
+			retval = g_mime_message_get_body (m, TRUE, &is_html);
+			g_mime_object_unref (GMIME_OBJECT(m));
+			g_free (s);
+		}
 	}
 
 	return retval;
 }
 
+gboolean
+article_get_header_run_hash_from_text (const gchar    * article_text,
+                                       GHashTable    ** makeme_hash,
+                                       gpointer       * allocme_buf)
+{
+	guint line_qty;
+	guint line_len;
+	const gchar * line; 
+	const gchar * march;
+	GHashTable * hash;
+	Run * buf;
+	guint buf_qty;
+	guint buf_size;
+
+	/* sanity clause */
+	g_return_val_if_fail (article_text, FALSE);
+	g_return_val_if_fail (makeme_hash!=NULL, FALSE);
+	g_return_val_if_fail (allocme_buf!=NULL, FALSE);
+
+	/* count number of lines */
+	line = NULL;
+	line_qty = 0;
+	for (march=article_text; get_next_token_run(march,'\n',&march,&line,&line_len) && line_len!=0; )
+		++line_qty;
+
+	/* alloc a buffer of enough Run structs */
+	buf_size = line_qty * 2u;
+	buf_qty = 0u;
+	buf = g_malloc (sizeof(Run) * buf_size);
+
+	/* alloc the hashtable */
+	hash = g_hash_table_new (run_hash, run_equal);
+
+	/* populate the hashtable */
+	march = article_text;
+	while (get_next_token_run (march, '\n', &march, &line, &line_len))
+	{
+		guint key_len = 0;
+		const gchar * key = NULL;
+
+		if (line_len == 0) /* \n\n -- end of headers */
+			break;
+
+		if (get_next_token_run (line, ':', NULL, &key, &key_len))
+		{
+			gint val_len = 0;
+			const gchar * val = NULL;
+			Run * key_run = buf + buf_qty++;
+			Run * val_run = buf + buf_qty++;
+
+			val = key + key_len + 1; /* skip past ':' */
+			val_len = 0;
+			while (val[val_len]!='\n' && val[val_len]!='\0')
+				++val_len;
+
+			g_hash_table_insert (hash, 
+			                     run_strstrip (run_init (key_run, key, key_len)),
+			                     run_strstrip (run_init (val_run, val, val_len)));
+		}
+	}
+
+	*makeme_hash = hash;
+	*allocme_buf = buf;
+	return TRUE;
+}
+
+
 static void
 article_set_author_from_header_impl (Article       * a,
                                      const gchar   * header_from,
                                      gboolean        init)
 {
+	gchar addr_buf[512];
+	gchar real_buf[512];
+	gchar * addr = addr_buf;
+	gchar * real = real_buf;
 	gchar * pch;
-	gchar * addr = NULL;
-	gchar * real = NULL;
 
 	g_return_if_fail (a!=NULL);
 	g_return_if_fail (a->group!=NULL);
@@ -1008,11 +1194,14 @@ article_set_author_from_header_impl (Article       * a,
 
 	/* note that strict is FALSE here because we can't control
 	 * how other people have posted to usenet */
-	gnksa_do_check_from (header_from, &addr, &real, FALSE);
-	if (pan_header_is_8bit_encoded (addr))
-		replace_gstr (&addr, g_mime_utils_8bit_header_decode((const guchar *)addr));
-	if (pan_header_is_8bit_encoded (real))
-		replace_gstr (&real, g_mime_utils_8bit_header_decode((const guchar *)real));
+	gnksa_do_check_from (header_from,
+	                     addr_buf, sizeof(addr_buf),
+	                     real_buf, sizeof(real_buf),
+	                     FALSE);
+	if (pan_header_is_8bit_encoded (addr_buf))
+		addr = g_mime_utils_8bit_header_decode ((const guchar *)addr_buf);
+	if (pan_header_is_8bit_encoded (real_buf))
+		real = g_mime_utils_8bit_header_decode ((const guchar *)real_buf);
 
 	/* use the real mail address, or fill in a default */
 	pch = addr;
@@ -1026,9 +1215,9 @@ article_set_author_from_header_impl (Article       * a,
 		a->author_real = group_chunk_string (a->group, real, TRUE);
 	}
 	else if (addr!=NULL && ((pch=strchr(addr,'@'))!=NULL)) {
-		gchar * tmp = g_strndup (addr, pch-addr);
+		gchar * tmp;
+		pan_strndup_alloca (tmp, addr, pch-addr);
 		a->author_real = group_chunk_string (a->group, tmp, TRUE);
-		g_free (tmp);
 	}
 	else {
 		const gchar * cpch = (gchar*)default_incoming_name_real;
@@ -1039,8 +1228,10 @@ article_set_author_from_header_impl (Article       * a,
 		articles_set_dirty (&a, 1);
 
 	/* cleanup */
-	g_free (addr);
-	g_free (real);
+	if (addr!=addr_buf)
+		g_free (addr);
+	if (real!=real_buf)
+		g_free (real);
 }
 
 void
@@ -1122,7 +1313,7 @@ article_set_from_raw_message   (Article         * a,
 	}
 
 	/* state */
-	pch = article_get_header (a, "Status");
+	pch = article_get_extra_header (a, "Status");
 	if (pch != NULL) {
 		const gboolean read = strchr (pch, 'R') != NULL;
 		articles_set_read (&a, 1, read);
@@ -1130,12 +1321,12 @@ article_set_from_raw_message   (Article         * a,
 	}
 
 	/* references */
-	pch = article_get_header (a, HEADER_REFERENCES);
+	pch = article_get_extra_header (a, HEADER_REFERENCES);
 	if (pch != NULL)
 		a->references = pch;
 
 	/* cleanup */
-	g_mime_message_destroy (msg);
+	g_mime_object_unref (GMIME_OBJECT(msg));
 	debug_exit ("article_set_from_raw_message");
 }
 
@@ -1143,7 +1334,7 @@ article_set_from_raw_message   (Article         * a,
 static gboolean
 breakpoint (const Article * a)
 {
-	if (a->group!=NULL)
+	if (a!=NULL && a->group!=NULL)
 		g_message ("bad article's group: [%s]", a->group->name);
 	return FALSE;
 }
@@ -1151,6 +1342,7 @@ breakpoint (const Article * a)
 gboolean
 article_is_valid (const Article * a)
 {
+	g_return_val_if_fail (a!=NULL, breakpoint(a));
 	g_return_val_if_fail (a->group!=NULL, breakpoint(a));
 	g_return_val_if_fail (group_is_valid(a->group), breakpoint(a));
 	g_return_val_if_fail (a->group->_articles_refcount>=0, breakpoint(a));
@@ -1218,39 +1410,6 @@ article_get_combined_linecount (const Article * a)
 	return retval;
 }
 
-/* Worker function for easily retrieving the newsreader */
-gchar*       
-article_get_newsreader (const Article * article)
-{
-	static gchar * tags[] = { HEADER_USER_AGENT": ",
-		                  HEADER_X_NEWSREADER": ",
-				  HEADER_X_MAILER": " };
-	static gint   tag_count = sizeof tags / sizeof tags[0];
-
-	gchar       * newsreader = NULL;
-	gchar       * headers = article_get_headers (article);
-	gchar      ** lines = g_strsplit (headers, "\n", -1);
-	gint          i;
-
-	for (i=0; newsreader == NULL && lines!=NULL && lines[i]!=NULL; ++i) {
-		gint tag;
-
-		for (tag = 0; tag < tag_count; tag++) {
-			gint tag_size = strlen(tags[tag]);
-
-			if (strncmp(lines[i], tags[tag], tag_size) == 0) {
-				newsreader = g_strdup (lines[i] + tag_size);
-				g_strstrip (newsreader);
-			}
-		}
-	}
-
-	g_strfreev (lines);	
-	g_free (headers);
-
-	return newsreader;
-}
-
 
 /***
 ****  EVENTS
@@ -1276,7 +1435,7 @@ fire_articles_changed (Group              * group,
 	g_return_if_fail (group!=NULL);
 	g_return_if_fail (articles!=NULL);
 	g_return_if_fail (article_qty>0);
-	g_return_if_fail (type==ARTICLE_CHANGED_READ || type==ARTICLE_CHANGED_DIRTY);
+	g_return_if_fail (type==ARTICLE_CHANGED_READ || type==ARTICLE_CHANGED_DIRTY || type==ARTICLE_CHANGED_NEW);
 
 	e.group = group;
 	e.articles = articles;
