@@ -24,6 +24,7 @@
 #include <errno.h>
 #include <math.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <sys/time.h>
 #include <unistd.h>
@@ -50,7 +51,7 @@
 static ServerGroupsType
 file_grouplist_097_style_exists (const Server * server)
 {
-	gchar * fname;
+	gchar fname [PATH_MAX];
 	ServerGroupsType retval = 0;
 
 	/* sanity clause */
@@ -58,16 +59,14 @@ file_grouplist_097_style_exists (const Server * server)
 	g_return_val_if_fail (is_nonempty_string(data_dir), FALSE);
 
 	/* check for sub */
-	fname = g_strdup_printf ("%s%c%s_sub.idx", data_dir, G_DIR_SEPARATOR, server->name);
+	g_snprintf (fname, sizeof(fname), "%s%c%s_sub.idx", data_dir, G_DIR_SEPARATOR, server->name);
 	if (file_exists (fname))
 		retval |= SERVER_GROUPS_SUBSCRIBED;
-	g_free (fname);
 
 	/* check for unsub */
-	fname = g_strdup_printf ("%s%c%s_unsub.idx", data_dir, G_DIR_SEPARATOR, server->name);
+	g_snprintf (fname, sizeof(fname), "%s%c%s_unsub.idx", data_dir, G_DIR_SEPARATOR, server->name);
 	if (file_exists (fname))
 		retval |= SERVER_GROUPS_UNSUBSCRIBED;
-	g_free (fname);
 
 	/* return results */
 	return retval;
@@ -76,15 +75,14 @@ file_grouplist_097_style_exists (const Server * server)
 static gboolean
 file_grouplist_096_style_exists (const Server * server)
 {
+	gchar fname [PATH_MAX];
 	gboolean retval;
-	gchar * fname;
 
 	g_return_val_if_fail (server!=NULL, FALSE);
 	g_return_val_if_fail (is_nonempty_string(data_dir), FALSE);
        
-	fname = g_strdup_printf ("%s%c%s.idx", data_dir, G_DIR_SEPARATOR, server->name);
+	g_snprintf (fname, sizeof(fname), "%s%c%s.idx", data_dir, G_DIR_SEPARATOR, server->name);
        	retval = file_exists (fname);
-	g_free (fname);
 
 	return retval;
 }
@@ -135,7 +133,8 @@ read_group_data (Server        * server,
 	version = get_next_token_int (march_idx, '\n', &march_idx);
 	groups = g_ptr_array_new ();
 	if (version==1 || version==2 || version==3 || version==4 || 
-	    version==5 || version==6 || version==7 || version==8)
+	    version==5 || version==6 || version==7 || version==8 ||
+	    version==9 || version==10)
 	{
 		long i;
 		glong qty = get_next_token_long (march_idx, '\n', &march_idx);
@@ -199,15 +198,32 @@ read_group_data (Server        * server,
 					g->identity_name = g_strdup (dat_str+l);
 			}
 
-			g->flags            = (gint8)   get_next_token_int   (march_idx, '\n', &march_idx);
+			g->flags = (gint8)   get_next_token_int   (march_idx, '\n', &march_idx);
 
 			if (version>=8) {
 				g->filter_show = get_next_token_ulong (march_idx, '\n', &march_idx);
 			}
 
-			g->filter_bits      = (guint32) get_next_token_ulong (march_idx, '\n', &march_idx);
-			g->sort_style       = (gint8)   get_next_token_int   (march_idx, '\n', &march_idx);
-			g->article_low      = (gulong)  get_next_token_ulong (march_idx, '\n', &march_idx);
+			g->filter_bits = (guint32) get_next_token_ulong (march_idx, '\n', &march_idx);
+
+			/* old sort style */
+			if (version >= 10)
+				g->old_sort_style = get_next_token_int (march_idx, '\n', &march_idx);
+
+			g->new_sort_style = (gint8) get_next_token_int (march_idx, '\n', &march_idx);
+
+			/* compare sort styles */
+			if (version >= 10)
+				if (abs(g->old_sort_style) == abs(g->new_sort_style))
+					g->old_sort_style = 0;
+
+			g->article_low = (gulong) get_next_token_ulong (march_idx, '\n', &march_idx);
+
+			if (version>=9)
+				g->loaded_since_last_fetch = get_next_token_int (march_idx, '\n', &march_idx) != 0;
+			else
+				g->loaded_since_last_fetch = FALSE;
+
 			g->article_high     = (gulong)  get_next_token_ulong (march_idx, '\n', &march_idx);
 
 			/* filter_bits got widened from guint16 to guint32 in version 7 */
@@ -266,7 +282,7 @@ read_group_data (Server        * server,
 
 	if (groups->len != 0) {
 		GPtrArray * unused = g_ptr_array_new ();
-		server_add_groups (server, (Group**)groups->pdata, groups->len, NULL, unused);
+		server_init_groups (server, (Group**)groups->pdata, groups->len, NULL, unused);
 		pan_g_ptr_array_foreach (unused, (GFunc)pan_object_unref, NULL);
 		g_ptr_array_free (unused, TRUE);
 	}
@@ -287,23 +303,23 @@ file_grouplist_load_file (Server * server, StatusItem * status, const gchar * fi
 	/* sanity clause */
 	g_return_val_if_fail (server!=NULL, 0);
 	g_return_val_if_fail (is_nonempty_string(server->name), 0);
-	g_return_val_if_fail (is_nonempty_string(filemarker), 0);
+	g_return_val_if_fail (filemarker!=NULL, 0);
 	g_return_val_if_fail (is_nonempty_string(data_dir), 0);
 
 	/* open the index file */
 	if (1) {
-		gchar * path = g_strdup_printf ("%s%c%s%s.idx", data_dir, G_DIR_SEPARATOR, server->name, filemarker);
+		gchar path [PATH_MAX];
+		g_snprintf (path, sizeof(path), "%s%c%s%s.idx", data_dir, G_DIR_SEPARATOR, server->name, filemarker);
 		pan_normalize_filename_inplace (path);
 		idx = read_file (path);
-		g_free (path);
 	}
 
 	/* open the data file */
 	if (1) {
-		gchar * path = g_strdup_printf ("%s%c%s%s.dat", data_dir, G_DIR_SEPARATOR, server->name, filemarker);
+		gchar path [PATH_MAX];
+		g_snprintf (path, sizeof(path), "%s%c%s%s.dat", data_dir, G_DIR_SEPARATOR, server->name, filemarker);
 		pan_normalize_filename_inplace (path);
 		dat = read_file (path);
-		g_free (path);
 	}
 
 	/* read the groups */
@@ -388,18 +404,18 @@ write_groups (FILE * idx_fp, FILE * dat_fp, GPtrArray * groups, StatusItem * sta
 
 	/* Write the group information... */
 	pos = 0;
-	fprintf (idx_fp, "8\n"); /* file format version number */
+	fprintf (idx_fp, "10\n"); /* file format version number */
 	fprintf (idx_fp, "%u\n", groups->len); /* number of groups */
 	for (i=0; i!=groups->len; ++i)
 	{
 		Group * group = GROUP(g_ptr_array_index (groups, i));
 
 		gchar * pch = group->_newsrc==NULL
-			? g_strdup("")
+			? NULL
 			: newsrc_get_read_str (group_get_newsrc(group));
 
 		gchar * purged = group->_purged==NULL
-			? g_strdup("")
+			? NULL
 			: newsrc_get_read_str (group_get_purged(group));
 
 		/* write the group name */
@@ -407,20 +423,22 @@ write_groups (FILE * idx_fp, FILE * dat_fp, GPtrArray * groups, StatusItem * sta
 		const long read_idx = get_string_offset (NULL, NULL, group->readable_name, dat_fp, &pos);
 		const long desc_idx = get_string_offset (NULL, NULL, group->description, dat_fp, &pos);
 		const long path_idx = get_string_offset (NULL, NULL, group->download_dir, dat_fp, &pos);
-		const long news_idx = get_string_offset (NULL, NULL, pch, dat_fp, &pos);
-		const long dead_idx = get_string_offset (NULL, NULL, purged, dat_fp, &pos);
+		const long news_idx = get_string_offset (NULL, NULL, pch?pch:"", dat_fp, &pos);
+		const long dead_idx = get_string_offset (NULL, NULL, purged?purged:"", dat_fp, &pos);
 		const long filt_idx = get_string_offset (NULL, NULL, group->filter_name, dat_fp, &pos);
 		const long iden_idx = get_string_offset (NULL, NULL, group->identity_name, dat_fp, &pos);
 
 		/* write the non-string fields. */
 		fprintf (idx_fp,
 		         "%ld\n" "%ld\n" "%ld\n" "%ld\n" "%ld\n" "%ld\n" "%ld\n" "%ld\n"
-		         "%u\n" "%lu\n" "%lu\n" "%d\n" "%lu\n" "%lu\n" "%lu\n" "%d\n" "%d\n" "%c\n",
+		         "%u\n" "%lu\n" "%lu\n" "%d\n" "%d\n" "%d\n" "%lu\n" "%lu\n" "%lu\n" "%d\n" "%d\n" "%c\n",
 		         name_idx, read_idx, desc_idx, path_idx, news_idx, dead_idx, filt_idx, iden_idx,
 		         (unsigned)(group->flags & ~GROUP_NEW),
 		         group->filter_show,
 		         (gulong)group->filter_bits,
-		         (int)group->sort_style,
+		         (int)group->old_sort_style,
+		         (int)group->new_sort_style,
+		         (int)(group->loaded_since_last_fetch ? 1 : 0),
 		         (unsigned long)group->article_low,
 		         (unsigned long)group->article_high,
 			 (unsigned long)group->article_high_old,
@@ -438,8 +456,10 @@ write_groups (FILE * idx_fp, FILE * dat_fp, GPtrArray * groups, StatusItem * sta
 		}
 
 		/* cleanup */
-		g_free (pch);
-		g_free (purged);
+		if (pch != NULL)
+			g_free (pch);
+		if (purged != NULL)
+			g_free (purged);
 	}
 
 	debug_exit ("write_groups");
@@ -459,17 +479,15 @@ save_groups_to_file (Server * server, GPtrArray * groups, gchar * filemarker, St
 
 	if (groups->len == 0)
 	{
-		gchar * pch;
-		pch = g_strdup_printf ("%s%c%s_%s.idx", data_dir, G_DIR_SEPARATOR, server->name, filemarker);
-		unlink (pch);
-		g_free (pch);
-		pch = g_strdup_printf ("%s%c%s_%s.dat", data_dir, G_DIR_SEPARATOR, server->name, filemarker);
-		unlink (pch);
-		g_free (pch);
+		gchar path [PATH_MAX];
+		g_snprintf (path, PATH_MAX, "%s%c%s_%s.idx", data_dir, G_DIR_SEPARATOR, server->name, filemarker);
+		unlink (path);
+		g_snprintf (path, PATH_MAX, "%s%c%s_%s.dat", data_dir, G_DIR_SEPARATOR, server->name, filemarker);
+		unlink (path);
 	}
 	else
 	{
-		gchar * pch;
+		gchar path [PATH_MAX];
 		FILE * idx_fp = NULL;
 		FILE * dat_fp = NULL;
 		gchar * idx_path = pan_make_temp (&idx_fp);
@@ -491,16 +509,14 @@ save_groups_to_file (Server * server, GPtrArray * groups, gchar * filemarker, St
 		fclose (dat_fp);
 
 		/* move the tmp files to the real thing */
-		pch = g_strdup_printf ("%s%c%s_%s.idx", data_dir, G_DIR_SEPARATOR, server->name, filemarker);
-		pan_normalize_filename_inplace (pch);
-		pan_rename (idx_path, pch);
+		g_snprintf (path, PATH_MAX, "%s%c%s_%s.idx", data_dir, G_DIR_SEPARATOR, server->name, filemarker);
+		pan_normalize_filename_inplace (path);
+		pan_rename (idx_path, path);
 		g_free (idx_path);
-		g_free (pch);
-		pch = g_strdup_printf ("%s%c%s_%s.dat", data_dir, G_DIR_SEPARATOR, server->name, filemarker);
-		pan_normalize_filename_inplace (pch);
-		pan_rename (dat_path, pch);
+		g_snprintf (path, PATH_MAX, "%s%c%s_%s.dat", data_dir, G_DIR_SEPARATOR, server->name, filemarker);
+		pan_normalize_filename_inplace (path);
+		pan_rename (dat_path, path);
 		g_free (dat_path);
-		g_free (pch);
 		debug_exit ("save_groups_to_file");
 	}
 }
