@@ -26,6 +26,8 @@
 #include <stdio.h>
 #include <glib.h>
 
+#include <gmime/memchunk.h>
+
 #include <pan/base/group.h>
 #include <pan/base/status-item.h>
 
@@ -37,6 +39,6 @@ void     file_headers_save             (Group * group, StatusItem * status);
 
 void     file_headers_save_noref       (Group * group, StatusItem * status);
 
-long     get_string_offset             (GHashTable *, GMemChunk *, const gchar *, FILE *, glong *);
+long     get_string_offset             (GHashTable *, MemChunk *, const gchar *, FILE *, glong *);
 
 #endif
