@@ -25,6 +25,20 @@
 
 #include <stddef.h>
 #include <glib.h>
+extern void* alloca (size_t);
+
+/**
+ * @param pch the string to duplicate
+ * lvalue is the duplicated string
+ */
+#define pan_strdup_alloca(pch) memcpy((char*)alloca(strlen(pch)+1),pch,strlen(pch)+1) 
+
+/**
+ * @param pch the char* to hold the duplicated string
+ * @param str the string to duplicate
+ * @param len the length of the string to duplicate
+ */
+#define pan_strndup_alloca(pch,str,len) ((pch)=(char*)alloca((len)+1),memcpy((pch),str,len),(pch)[len]='\0',(pch))
 
 /**
 ***  Strings
@@ -32,24 +46,24 @@
 
 #define is_nonempty_string(A) ((A) && *(A)!='\0')
 
-gint pan_strcmp (const gchar* a, const gchar  * b);
+int pan_strcmp (const char* a, const char  * b);
 
-gchar* pan_stristr (const gchar * string, const gchar * pattern);
+char* pan_stristr (const char * string, const char * pattern);
 
-gboolean string_ends_with (const gchar* str, const gchar* end);
+gboolean string_ends_with (const char* str, const char* end);
 
-void replace_gstr (gchar ** target_free_old, gchar * assign_from_me);
+void replace_gstr (char ** target_free_old, char * assign_from_me);
 
-gchar* pan_substitute (const gchar  * original,
-                       const gchar  * search,
-                       const gchar  * replace);
+char* pan_substitute (const char  * original,
+                      const char  * search,
+                      const char  * replace);
 
 /**
 ***  Glib
 **/
 
-gchar* pan_strstr (const gchar * s1,
-                   const gchar * s2);
+char* pan_strstr (const char * s1,
+                  const char * s2);
 
 void pan_g_ptr_array_append (GPtrArray   * target,
                              gpointer    * source_ptr,
@@ -82,9 +96,9 @@ void pan_g_string_replace (GString    * str,
 
 void pan_g_string_strstrip (GString    * str);
 
-void pan_g_string_append_len (GString       * str,
-                              const gchar   * appendme,
-                              guint           len);
+void pan_g_string_append_len (GString      * str,
+                              const char   * appendme,
+                              guint          len);
 
 /**
 ***  Tokens
@@ -99,9 +113,9 @@ int get_next_token_int (const char   * pch,
                         const char     delimiter,
                         const char  ** setme_next_token);
 
-glong get_next_token_long (const char   * pch,
-                           const char     delimiter,
-                           const char  ** setme_next_token);
+long get_next_token_long (const char   * pch,
+                          const char     delimiter,
+                          const char  ** setme_next_token);
 
 gulong get_next_token_ulong (const char   * pch,
                              const char     delimiter,
@@ -116,17 +130,19 @@ gboolean get_next_token_g_str (const char   * pch,
                                const char  ** setme_next_token,
                                GString      * setme);
 
-const gchar*  get_next_token_run (const gchar   * pch,
-                                  const gchar     delimiter,
-                                  const gchar  ** setme_next_token,
-                                  const gchar  ** setme_start,
-                                  gint          * setme_len);
+const char*  get_next_token_run (const char   * pch,
+                                 const char     delimiter,
+                                 const char  ** setme_next_token,
+                                 const char  ** setme_start,
+                                 int          * setme_len);
 
 
-gchar* pan_str_escape   (const gchar * encode_for_xml);
+char* pan_str_escape   (const char * encode_for_xml);
 
-gchar* pan_str_unescape (const gchar * decode_xml);
+char* pan_str_unescape (const char * decode_xml);
 
+char** strsplit_command_line_arguments (const char * command_line);
+  
 
 /**
  * Formats a number with commas; ie, 10321 becomes "10,321"
