@@ -26,6 +26,8 @@
 #include <time.h>
 #include <glib.h>
 
+#include <gmime/memchunk.h>
+
 #include <pan/base/pan-callback.h>
 #include <pan/base/pan-object.h>
 #include <pan/base/status-item.h>
@@ -64,7 +66,7 @@ typedef struct
 
 	ServerGroupsType _groups_dirty;
 	ServerGroupsType _groups_loaded;
-	GMemChunk * _group_memchunk;
+	MemChunk * _group_memchunk;
 	GPtrArray * _groups;             /* don't access directly; use server_get_groups instead */
 }
 Server;
@@ -108,6 +110,12 @@ struct _Group* server_alloc_new_group    (Server            * server);
 struct _Group* server_get_named_group    (Server            * server,
                                           const gchar       * name);
 
+void       server_init_groups             (Server            * server,
+                                          struct _Group    ** groups,
+                                          int                 group_qty,
+                                          GPtrArray         * fillme_used,
+                                          GPtrArray         * fillme_not_used);
+
 void       server_add_groups             (Server            * server,
                                           struct _Group    ** groups,
                                           int                 group_qty,
