@@ -138,10 +138,10 @@ uu_get_char_len (gint octet_len)
 	gint char_len = (octet_len / 3) * 4;
 	switch (octet_len % 3) {
 		case 0: break;
-		case 1: char_len += 2;
-		case 2: char_len += 3;
+		case 1: char_len += 2; break;
+		case 2: char_len += 3; break;
 	}
-	return octet_len;
+	return char_len;
 }
 
 gboolean
@@ -169,6 +169,12 @@ is_uu_line (const gchar * line, gint len)
 	if (char_len+1 > len)
 		return FALSE;
 
+	/* if there's a lot of noise at the end, be suspicious.
+	   This is to keep out lines of nntp-server-generated
+	   taglines that have a space as the first character */
+	if (char_len+10 < len)
+		return FALSE;
+
 	/* make sure each character is in the uuencoded range */
 	for (i=0; i<char_len; ++i)
 		if (line[i+1]<0x20 || line[i+1]>0x60)
@@ -393,6 +399,8 @@ look_for_uuencoded_data (GMimePart * parent, gpointer data)
 
 				g_hash_table_insert (skip, child, child);
 				g_mime_part_add_subpart (parent, child);
+
+				g_mime_object_unref (GMIME_OBJECT (child));
 			}
 		}
 
@@ -418,6 +426,8 @@ construct_message_impl (GMimeStream * istream)
 	GMimeMessage * message;
 	GHashTable * skip;
 
+	g_return_val_if_fail (istream!=NULL, NULL);
+
 	message = g_mime_parser_construct_message (istream, TRUE);
 
 	/* walk through all the parts to break uu from non-uu */
@@ -434,6 +444,8 @@ pan_g_mime_parser_construct_message (const gchar * data)
 	GMimeStream * istream;
 	GMimeMessage * retval;
 
+	g_return_val_if_fail (is_nonempty_string(data), NULL);
+
 	istream = g_mime_stream_mem_new_with_buffer (data, strlen(data));
 	retval = construct_message_impl (istream);
 	g_mime_stream_unref (istream);
@@ -448,6 +460,8 @@ pan_g_mime_parser_construct_message_from_file (FILE * fp)
 	GMimeStream * buf_istream;
 	GMimeMessage * message;
 
+	g_return_val_if_fail (fp!=NULL, NULL);
+
 	istream = g_mime_stream_file_new (fp);
 	buf_istream = g_mime_stream_buffer_new (istream, GMIME_STREAM_BUFFER_BLOCK_READ);
        	message = construct_message_impl (buf_istream);
