@@ -238,7 +238,8 @@ fill_body (const gchar * body, gint wrap_column)
 			g_string_append (s, (const char*)g_ptr_array_index(array,i));
 			g_string_append_c (s, '\n');
 		}
-		g_string_truncate (s, s->len-1);
+		if (s->len > 0)
+			g_string_truncate (s, s->len-1);
 		new_body = s->str;
 		g_string_free (s, FALSE);
 	}
@@ -305,9 +306,8 @@ mute_quoted_text_in_body (const gchar * text)
 
 
 gchar*
-rot13 (const gchar * text)
+rot13_inplace (gchar * text)
 {
-	gchar * retval = NULL;
 	const char * plain = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
 	const char * roted = "nopqrstuvwxyzabcdefghijklmNOPQRSTUVWXYZABCDEFGHIJKLM";
 
@@ -315,9 +315,7 @@ rot13 (const gchar * text)
 	{
 		gchar * march;
 
-		retval = g_strdup (text);
-
-		for (march=retval; *march!='\0'; ++march)
+		for (march=text; *march!='\0'; ++march)
 		{
 			const gchar * offset = strchr (plain, *march);
 			if (offset != NULL)
@@ -328,5 +326,5 @@ rot13 (const gchar * text)
 		}
 	}
 
-	return retval;
+	return text;
 }
