@@ -19,7 +19,8 @@
 #include <config.h>
 #include <string.h>
 #include <gnome.h>
-#include "pan-about.h"
+#include <pan/base/pan-i18n.h>
+#include <pan/dialogs/pan-about.h>
 
 typedef struct
 {
