@@ -3,7 +3,7 @@
 
 #include <pan/base/group.h>
 
-void group_empty_dialog (Group *group);
+void group_empty_dialog (const GPtrArray * groups);
 
 void group_new_folder_dialog (void);
 
