@@ -94,16 +94,14 @@ gui_notebook_construct (void)
 }
 
 void
-gui_notebook_page_set (int page, GtkWidget * focus_item)
+gui_notebook_page_set_nolock (int page, GtkWidget * focus_item)
 {
-	pan_lock();
 	gtk_notebook_set_page (GTK_NOTEBOOK (notebook), page);
 	gtk_widget_grab_focus (focus_item);
-	pan_unlock();
 }
 
 int
-gui_notebook_get_current_pane (void)
+gui_notebook_get_current_pane_nolock (void)
 {
 	return gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));
 }
