@@ -136,6 +136,15 @@ write_identity (GString        * gstr,
 		g_string_append (gstr, "</attribution>");
 	}
 
+	/* msg_id fqdn */
+	if (is_nonempty_string(id->msg_id_fqdn))
+	{
+		newline_depth (gstr, depth);
+		g_string_append (gstr, "<msg_id_fqdn>");
+		pan_g_string_append_escaped (gstr, id->msg_id_fqdn);
+		g_string_append (gstr, "</msg_id_fqdn>");
+	}
+
 	/* custom headers */
 	if (id->custom_headers->len > 0)
 	{
@@ -218,23 +227,28 @@ identity_xml_to_string (Identity    ** identities,
 	return retval;
 }
 
-static void
+static gboolean
 identity_xml_fwrite (FILE         * fp,
                      Identity    ** identities,
                      gint           identities_qty,
                      const gchar  * default_news_id,
                      const gchar  * default_mail_id)
 {
-        gchar * pch;
+        gchar  * pch;
+	size_t   to_write, written;
 
-        g_return_if_fail (fp!=NULL);
-        g_return_if_fail (identities!=NULL);
-        g_return_if_fail (identities_qty>0);
+        g_return_val_if_fail (fp!=NULL, 0);
+        g_return_val_if_fail (identities!=NULL, 0);
+        g_return_val_if_fail (identities_qty>0, 0);
 
         pch = identity_xml_to_string ((Identity**)identities, identities_qty,
 		default_news_id, default_mail_id);
-        fwrite (pch, sizeof(char), strlen(pch), fp);
+
+	to_write = strlen (pch);
+        written = fwrite (pch, sizeof(char), to_write, fp);
         g_free (pch);
+
+	return to_write == written;
 }
 
 void
@@ -250,25 +264,34 @@ identity_xml_write (const gchar  * filename,
 
         /* write the file */
 
-        if (identities_qty == 0)
-        {
+        if (identities_qty == 0) {
                 unlink (filename);
         }
         else
         {
-                FILE * fp = fopen (filename, "w+");
-
-                if (fp == NULL) {
-                        log_add_va (LOG_ERROR,
-                                _("Can't write profiles to `%s': %s"),
-                                filename, g_strerror(errno));
-                } else {
-                        identity_xml_fwrite (fp, identities, identities_qty,
-				default_news_id, default_mail_id);
+		FILE     * fp;
+		gchar    * tmp_path = pan_make_temp (&fp);
+		gboolean   ok = fp ? TRUE : FALSE;
+	      
+                if (ok) {	
+			ok = identity_xml_fwrite (fp, identities, identities_qty, default_news_id, default_mail_id);
                         fclose (fp);
+                }
+                
+		if (ok) {
+			ok = pan_rename (tmp_path, filename); 
+		}
+
+		if (ok) {
                         log_add_va (LOG_INFO, _("Wrote %d profiles to '%s'"),
                                 identities_qty, filename);
-                }
+		}
+		else {
+			log_add_va (LOG_ERROR, _("Can't write profiles to `%s': %s"),
+                                filename, g_strerror(errno));
+		}
+
+		g_free (tmp_path);
         }
 }
 
@@ -391,6 +414,9 @@ create_identity (xmlDoc *doc, xmlNode *node)
 			else
 			if (!pan_strcmp (name, "attribution"))
 				fldptr = &id->attribution;
+			else
+			if (!pan_strcmp (name, "msg_id_fqdn"))
+				fldptr = &id->msg_id_fqdn;
 
 			if (fldptr != NULL && child->xmlChildrenNode != NULL)
 			{
