@@ -39,6 +39,7 @@
 #include <pan/nntp.h>
 #include <pan/prefs.h>
 #include <pan/task.h>
+#include <pan/auth-spa.h>
 
 gboolean online_prompt = TRUE;
 
@@ -56,6 +57,11 @@ enum
 	ERR_ACCESS				= 502
 };
 
+enum
+{
+	AUTH_UNKNOWN,
+	AUTH_MSN
+};
 
 /**
 ***
@@ -239,11 +245,87 @@ nntp_handshake (StatusItem  * status,
 	return val;
 }
 
+static gint
+nntp_spa_authenticate (StatusItem     * status,
+                        PanSocket      * sock)
+{
+	/* Authentication for SPA scheme (msn.com, and others) */
+	/* Added by Marc Prud'hommeaux <marc@apocalypse.org>, mostly */
+	/* taken from the ftechmail SPA authentication, which, in */
+	/* turn, was ripped out of the SAMBA project. */
+
+	gint val;
+	const gchar * response = NULL;
+	const gchar * response_trim = NULL;
+	gchar msgbuf[2048];
+	SPAAuthRequest   spa_request;
+	SPAAuthChallenge spa_challenge;
+	SPAAuthResponse  spa_response;
+
+	debug_enter ("nntp_spa_authenticate");
+
+	memset (msgbuf, 0, sizeof (msgbuf));
+
+	if (nntp_command (status, sock, &response, "AUTHINFO GENERIC MSN")
+	!= NEED_AUTHDATA) {
+		status_item_emit_error_va (status, _(
+			"Authentication failed: bad handshake for SPA password"));
+		return TASK_FAIL;
+	}
+	
+	spa_build_auth_request (&spa_request, sock->nntp_username, NULL);
+	spa_bits_to_base64 (msgbuf, (unsigned char*)&spa_request, 
+		spa_request_length (&spa_request));
+
+	val = nntp_command_va (status, sock, &response, 
+		"AUTHINFO GENERIC %s", msgbuf);
+	if (val != NEED_AUTHDATA) {
+		status_item_emit_error_va (status, _(
+			"Bad SPA handshake: %s"), response);
+		debug_exit ("nntp_spa_authenticate");
+		return TASK_FAIL;
+	}
+
+	if (!is_nonempty_string(sock->nntp_password)) {
+		if (status != NULL)
+			status_item_emit_error_va (status, _(
+				"Authentication failed: need a password"));
+		return TASK_FAIL;
+	}
+
+	/* response is now something like:					*/
+	/* 381 TlRMTVNTUAACAAAAGAAYACAAAAABAg [...]			*/
+	/* Trim off the first 4 bytes in our response.		*/
+	response_trim = response + 4;
+	spa_base64_to_bits ((unsigned char*)&spa_challenge, response_trim);
+
+	spa_build_auth_response (&spa_challenge, &spa_response, 
+		sock->nntp_username, sock->nntp_password);
+
+	spa_bits_to_base64 (msgbuf, (unsigned char*)&spa_response, 
+		spa_request_length (&spa_response));
+
+	val = nntp_command_va (status, sock, &response, 
+		"AUTHINFO GENERIC %s", msgbuf);
+
+	if (val != OK_AUTH) {
+		status_item_emit_error_va (status, _(
+			"Authentication failed: %s"), response);
+		debug_exit ("nntp_spa_authenticate");
+		return TASK_FAIL;
+	}
+
+	debug_exit ("nntp_spa_authenticate");
+	return TASK_SUCCESS;
+}
+
 gint
-nntp_authenticate (StatusItem     * status,
-                   PanSocket      * sock)
+nntp_authenticate_types (StatusItem     * status,
+                         PanSocket      * sock,
+                         gboolean         check_auth_types)
 {
 	gint val;
+	int authtype = AUTH_UNKNOWN;
 	const gchar * response = NULL;
 	debug_enter ("nntp_authenticate");
 
@@ -252,8 +334,54 @@ nntp_authenticate (StatusItem     * status,
 	g_return_val_if_fail (sock!=NULL, TASK_FAIL);
 	g_return_val_if_fail (sock->nntp_username!=NULL, TASK_FAIL_HOPELESS);
 
+	/* check for authentication schemes */
+	/* see: http://www.mibsoftware.com/userkt/nntpext/0035.htm */
+	if (check_auth_types) {
+		val = nntp_command_va (status, sock, &response, "AUTHINFO GENERIC");
+		if (val==OK_AUTH) {
+			/* now we should be getting a list of valid authentication */
+			/* types terminated with a "." */
+			while (nntp_get_response (sock, &response) != -9999
+			&& response != NULL
+			&& pan_strcmp (response, ".\r\n")) {
+				/* supported authentication schemes follow */
+				if (!pan_strcmp (response, "MSN\r\n")) {
+					authtype = AUTH_MSN;
+				}
+				else {
+					status_item_emit_error_va (status, _(
+						"Unsupported authentication mechanism: %s"), response);
+				}
+			}
+	
+			switch (authtype)
+			{
+				case AUTH_MSN:
+					return nntp_spa_authenticate (status, sock);
+	
+				/* Unknown authentication type. We will fall through	*/
+				/* and try the normal authentication method anyway.		*/
+				case AUTH_UNKNOWN:
+				default:
+					status_item_emit_error_va (status, _(
+						"No supported authentication mechanism"));
+			}
+		}
+	}
+
 	/* set the username */
 	val = nntp_command_va (status, sock, &response, "AUTHINFO USER %s", sock->nntp_username);
+
+	/* On servers that require SPA authentication, sending			*/
+	/* "AUTHINFO USER ..." seems to result in a 502 error. If we	*/
+	/* have not yet tried other authentication methods, do so now.	*/
+	/* Ideally, we would not need to do this, but since some news`	*/
+	/* servers disconnect the user if they send "AUTHINFO GENERIC",	*/
+	/* we need to first try simple authentication.					*/
+	if (val==ERR_ACCESS && check_auth_types==FALSE) {
+		return nntp_authenticate_types (status, sock, TRUE);
+	}
+
 	if (val!=OK_AUTH && val!=NEED_AUTHDATA) {
 		if (status != NULL)
 			status_item_emit_error_va (status, _("Authentication failed: %s"), response);
@@ -281,6 +409,13 @@ nntp_authenticate (StatusItem     * status,
 	return TASK_SUCCESS;
 }
 
+gint
+nntp_authenticate (StatusItem     * status,
+                   PanSocket      * sock)
+{
+	return nntp_authenticate_types (status, sock, FALSE);
+}
+
 gint
 nntp_can_post (StatusItem   * status,
                PanSocket    * sock,
@@ -425,30 +560,30 @@ nntp_noop (StatusItem    * status,
 
 static gboolean
 init_header_if_match (Article       * a,
+                      const gchar   * token,
+                      const gchar   * token_key,
+                      guint           token_key_len,
                       const gchar   * header,
-                      const gchar   * header_key,
-                      const gchar   * article_key,
                       HeaderAction    action)
 {
 	gboolean match = FALSE;
-	size_t len;
 
 	/* entry assertions */
 	g_return_val_if_fail (a!=NULL, FALSE);
-	g_return_val_if_fail (is_nonempty_string(header_key), FALSE);
-	g_return_val_if_fail (is_nonempty_string(article_key), FALSE);
-	g_return_val_if_fail (header!=NULL, FALSE);
+	g_return_val_if_fail (is_nonempty_string(token_key), FALSE);
+	g_return_val_if_fail (token_key_len>0, FALSE);
+	g_return_val_if_fail (is_nonempty_string(header), FALSE);
+	g_return_val_if_fail (token!=NULL, FALSE);
 
-	/* if this is the header we're looking for, keep it. */
-	len = strlen (header_key);
-	if (!strncmp (header, header_key, len))
+	/* if this is the token we're looking for, keep it. */
+	if (!strncmp (token, token_key, token_key_len))
 	{
-		const gchar * val = header + len;
+		const gchar * val = token + token_key_len;
 		if (!pan_header_is_8bit_encoded (val))
-			article_init_header (a, article_key, val, action);
+			article_init_header (a, header, val, action);
 		else {
 			gchar * b = g_mime_utils_8bit_header_decode ((const guchar*)val);
-			article_init_header (a, article_key, b, action);
+			article_init_header (a, header, b, action);
 			g_free (b);
 		}
 		match = TRUE;
@@ -458,19 +593,19 @@ init_header_if_match (Article       * a,
 }
 
 gint
-nntp_article_download (StatusItem      * status,
-                       PanSocket       * sock,
-                       Group           * group,
-                       Article         * article,
-                       const gboolean  * abort,
-                       gint              verbose,
-                       gboolean          checkout)
+nntp_article_download_to_gstr (StatusItem      * status,
+                               PanSocket       * sock,
+                               Group           * group,
+                               Article         * article,
+                               const gboolean  * abort,
+                               gint              verbose,
+                               gboolean          checkout,
+                               GString         * setme)
 {
 	const gchar * response;
 	gboolean getting_headers;
 	gint retval;
 	gint val;
-	GString * buf;
 	GString * line;
 	debug_enter ("nntp_article_download");
 
@@ -505,7 +640,8 @@ nntp_article_download (StatusItem      * status,
 	/* try to read the article... */
 	getting_headers = TRUE;
 	line = g_string_sized_new (80);
-	buf = g_string_sized_new (80*400); /* prealloc space for ~400 lines */
+
+	g_string_truncate (setme, 0);
 	retval = TASK_SUCCESS;
 	if (verbose & NNTP_VERBOSE_INIT_STEPS)
 		status_item_emit_init_steps (status, article->linecount);
@@ -538,7 +674,7 @@ nntp_article_download (StatusItem      * status,
 			g_string_erase (line, 0, 1);
 
 		/* save the line */
-		g_string_append (buf, line->str);
+		g_string_append (setme, line->str);
 		if (verbose & NNTP_VERBOSE_NEXT_STEP)
 			status_item_emit_next_step (status);
 
@@ -549,50 +685,73 @@ nntp_article_download (StatusItem      * status,
 				getting_headers = FALSE;
 			else 
 			{
+				gint i;
 				Article * a = article;
 				gboolean matched = FALSE;
 
+				struct table {
+					const gchar * token;
+					guint token_len;
+					const gchar * header;
+					gint header_action;
+				};
+				static struct table tokens[] = {
+					{ "Newsgroups: ", 12, HEADER_NEWSGROUPS, DO_CHUNK_SHARE },
+					{ "Reply-To: ", 10, HEADER_REPLY_TO, DO_CHUNK_SHARE },
+					{ "Followup-To: ", 13, HEADER_FOLLOWUP_TO, DO_CHUNK_SHARE },
+					{ "Xref: ", 6, HEADER_XREF, DO_CHUNK },
+				};
+				static gint TOKEN_QTY = sizeof(tokens) / sizeof(tokens[0]);
+
 				/* trim linefeed */
 				g_strstrip (line->str);
 				line->len = strlen(line->str);
 
-				/* check for headers to store in Article */
-				if (!matched)
-					matched = init_header_if_match (a, line->str,
-					                                "Newsgroups: ",
-					                                HEADER_NEWSGROUPS,
-					                                DO_CHUNK_SHARE);
-				if (!matched)
-					matched = init_header_if_match (a, line->str,
-					                                "Reply-To: ",
-					                                HEADER_REPLY_TO,
-					                                DO_CHUNK_SHARE);
-				if (!matched)
-					matched = init_header_if_match (a, line->str,
-					                                "Followup-To: ",
-					                                HEADER_FOLLOWUP_TO,
-					                                DO_CHUNK_SHARE);
-				if (!matched)
-					matched = init_header_if_match (a, line->str,
-					                                "Xref: ",
-					                                HEADER_XREF,
-					                                DO_CHUNK);
+				/* check the header */
+				for (i=0; !matched && i<TOKEN_QTY; ++i)
+					matched = init_header_if_match (a,
+					                                line->str,
+					                                tokens[i].token,
+					                                tokens[i].token_len,
+					                                tokens[i].header,
+					                                tokens[i].header_action);
 			}
 		}
 	}
 
 	/* save the message */
 	if (retval == TASK_SUCCESS) {
-        	acache_set_message (article_get_message_id(article), buf->str, checkout);
+        	acache_set_message (article_get_message_id(article), setme->str, setme->len+1, checkout);
 		articles_remove_flag (&article, 1, STATE_DOWNLOAD_FLAGGED);
 	}
 
-	g_string_free (buf, TRUE);
 	g_string_free (line, TRUE);
 	debug_exit ("nntp_article_download");
 	return retval;
 }
 
+gint
+nntp_article_download (StatusItem      * status,
+                       PanSocket       * sock,
+                       Group           * group,
+                       Article         * article,
+                       const gboolean  * abort,
+                       gint              verbose,
+                       gboolean          checkout)
+{
+	GString * tmp;
+	gint retval;
+	debug_enter ("nntp_article_download");
+
+	/* download the message */
+ 	tmp = g_string_sized_new (80 * 400); /* preallocate space for ~400 lines */
+	retval = nntp_article_download_to_gstr (status, sock, group, article, abort, verbose, checkout, tmp);
+	g_string_free (tmp, TRUE);
+
+	debug_exit ("nntp_article_download");
+	return retval;
+}
+
 
 /**
 ***
@@ -725,6 +884,7 @@ nntp_download_headers (StatusItem       * status,
 
 		/* create the article data */
 		a = article_new (group);
+		a->is_new = TRUE;
 
 		/* setup for parsing */
 		g_string_assign (buf, buffer);
@@ -785,10 +945,11 @@ nntp_download_headers (StatusItem       * status,
 		/* get crossref */
 		if (get_next_token_run (march, '\t', &march, &tok, &tok_len)) {
 			((gchar*)tok)[tok_len] = '\0';
-			if (tok_len!=0 || strncmp(tok,"Xref: ", 6))
+			if (tok_len==0 || strncmp(tok,"Xref: ", 6))
 				a->crosspost_qty = 1;
 			else {
 				const gchar * foo;
+				g_strstrip ((gchar*)tok); /* remove trailing linefeed */
 				a->crosspost_qty = 0;
 				for (foo=tok+6; *foo; ++foo)
 					if (*foo==':')
@@ -910,13 +1071,41 @@ nntp_download_bodies (StatusItem       * status,
 	gint lines_total;
 	debug_enter ("nntp_download_bodies");
 
-	/* sanity checks */
+	/* sanity checks pt 1 */
+	g_return_val_if_fail (articles!=NULL, TASK_FAIL_HOPELESS);
+	g_return_val_if_fail (index!=NULL, TASK_FAIL_HOPELESS);
+
+	/**
+	***  If we have the bodies, then just unflag 'em and maybe check them out
+	**/
+
+	for (i=*index; i<articles->len; ++i)
+		if (!article_has_body (ARTICLE(g_ptr_array_index(articles,i))))
+			break;
+
+	if (i==articles->len) {
+		const gint qty = articles->len - *index;
+		if (qty > 0) {
+			articles_remove_flag ((Article**)articles->pdata+*index, qty, STATE_DOWNLOAD_FLAGGED);
+			if (checkout) {
+				const gchar ** message_ids = alloca (sizeof(gchar*) * qty);
+				for (i=*index; i<articles->len; ++i)
+					message_ids[i] = article_get_message_id (ARTICLE(g_ptr_array_index(articles,i)));
+				acache_file_checkout (message_ids, qty);
+			}
+		}
+		return TASK_SUCCESS;
+	}
+
+	/* sanity checks pt 2 */
 	g_return_val_if_fail (status!=NULL, TASK_FAIL_HOPELESS);
 	g_return_val_if_fail (group_is_valid(group), TASK_FAIL_HOPELESS);
 	g_return_val_if_fail (sock!=NULL, TASK_FAIL);
 	g_return_val_if_fail (abort!=NULL, TASK_FAIL_HOPELESS);
-	g_return_val_if_fail (articles!=NULL, TASK_FAIL_HOPELESS);
-	g_return_val_if_fail (index!=NULL, TASK_FAIL_HOPELESS);
+
+	/**
+	***  Phooey, we have to download some of the bodies
+	**/
 
 	/* change to the group we want */
 	if (*abort)
@@ -984,8 +1173,9 @@ nntp_cancel (StatusItem     * status,
              PanSocket      * sock)
 {
 	GString * msg;
-	gchar * author;
 	const gchar * newsgroups;
+	gchar author[512];
+	gchar * author_encoded;
 	gint retval;
 	const gchar * msg_id = article_get_message_id (article);
 	debug_enter ("nntp_cancel");
@@ -995,7 +1185,7 @@ nntp_cancel (StatusItem     * status,
 	g_return_val_if_fail (status != NULL, TASK_FAIL_HOPELESS);
 	g_return_val_if_fail (article_is_valid(article), TASK_FAIL_HOPELESS);
 	g_return_val_if_fail (is_nonempty_string(msg_id), TASK_FAIL_HOPELESS);
-	newsgroups = article_get_header (article, HEADER_NEWSGROUPS);
+	newsgroups = article_get_extra_header (article, HEADER_NEWSGROUPS);
 	g_return_val_if_fail (is_nonempty_string(newsgroups), TASK_FAIL_HOPELESS);
 
 	/* let the user know what we're doing */
@@ -1003,9 +1193,9 @@ nntp_cancel (StatusItem     * status,
 
 	/* build the message to post */
 	msg = g_string_new (NULL);
-	author = article_get_author_str(article);
-        replace_gstr (&author, g_mime_utils_8bit_header_encode((const guchar*)author));
-	g_string_sprintfa (msg, "From: %s\r\n", author);
+	article_get_author_str (article, author, sizeof(author));
+	author_encoded = g_mime_utils_8bit_header_encode ((const guchar*)author);
+	g_string_sprintfa (msg, "From: %s\r\n", author_encoded);
 	g_string_sprintfa (msg, "Newsgroups: %s\r\n", newsgroups);
 	g_string_sprintfa (msg, "Subject: cancel %s\r\n", msg_id);
 	g_string_sprintfa (msg, "Control: cancel %s\r\n", msg_id);
@@ -1016,7 +1206,7 @@ nntp_cancel (StatusItem     * status,
 	retval = nntp_post (status, sock, msg->str);
 
 	/* cleanup */
-	g_free (author);
+	g_free (author_encoded);
 	g_string_free (msg, TRUE);
 
 	debug_exit ("nntp_cancel");
