@@ -35,7 +35,7 @@
 #include <pan/base/pan-i18n.h>
 #include <pan/base/serverlist.h>
 
-#include <pan/filters/filter-ui.h>
+#include <pan/filter-ui.h>
 #include <pan/filters/filter-manager.h>
 #include <pan/filters/filter-top.h>
 
@@ -500,12 +500,11 @@ edit_filters_cb (GtkButton * button, gpointer data)
 	gtk_widget_show_all (w);
 }
 
-static gint
+static void
 filters_changed_cb (gpointer a, gpointer b, gpointer c)
 {
 	RuleEditDialog * d = (RuleEditDialog*)c;
 	populate_filter_tab (d, d->filter_name);
-	return 0;
 }
 
 static GtkWidget*
