@@ -39,10 +39,12 @@
 #include <pan/base/pan-i18n.h>
 #include <pan/base/status-item.h>
 
+#include <pan/identities/identity.h>
+#include <pan/identities/identity-manager.h>
+
 #include <pan/prefs.h>
 #include <pan/smtp.h>
 #include <pan/sockets.h>
-#include <pan/util.h>
 
 extern const gchar* sockread_err_msg;
 extern const gchar* sockwrite_err_msg;
@@ -180,7 +182,7 @@ smtp_send_article (StatusItem    * item,
 	/* sanity checks */
 	g_return_val_if_fail (item!=NULL, -1);
 	g_return_val_if_fail (sock!=NULL, -1);
-	g_return_val_if_fail (article!=NULL, -1);
+	g_return_val_if_fail (article_is_valid(article), -1);
 
 	status_item_emit_status (item, _("Preparing to Send Mail"));
 
@@ -189,7 +191,7 @@ smtp_send_article (StatusItem    * item,
 	**/
 
 	/* who is this mail to? */
-	mail_to = article_get_header (article, PAN_MAIL_TO);
+	mail_to = article_get_extra_header (article, PAN_MAIL_TO);
 	pan_warn_if_fail (is_nonempty_string (mail_to));
 	if (!is_nonempty_string(mail_to)) {
 		status_item_emit_error (item, _("No Recipients!"));
@@ -199,7 +201,7 @@ smtp_send_article (StatusItem    * item,
 	/* who is this mail from?  (part 1) */
 	reverse_path = NULL;
 	if (okay) {
-		reverse_path = article_get_header (article, PAN_REVERSE_PATH);
+		reverse_path = article_get_extra_header (article, PAN_REVERSE_PATH);
 		if (!is_nonempty_string(reverse_path)) {
 			g_message (_("Can't parse the From: line; hope for the best."));
 			reverse_path = article->author_addr;
@@ -213,16 +215,19 @@ smtp_send_article (StatusItem    * item,
 	/* who is this mail from?  (part 2) */
 	from = NULL;
 	if (okay) {
-		gchar * author = article_get_author_str (article);
+		gchar author[512];
+		article_get_author_str (article, author, sizeof(author));
 		if (is_nonempty_string (author))
 			from = g_strdup (author);
-		if (from == NULL)
-			from = get_default_author_address ();
+		if (from == NULL) {
+			Identity * id = identity_manager_get_default (ID_MAIL_DEFAULT);
+			if (id != NULL)
+				from = g_strdup (id->author_addr);
+		}
 		if (from == NULL) {
 			status_item_emit_error(item, _("No Sender Specified!"));
 			okay = FALSE;
 		}
-		g_free (author);
 	}
 
 	/**
@@ -253,14 +258,15 @@ smtp_send_article (StatusItem    * item,
 			{
 				const gchar * fmt = "RCPT TO: <%s>\r\n";
 				const char * name = (const char*) g_ptr_array_index (to_array, i);
-				gchar * address;
-				gchar * realname;
+				gchar address[512];
+				gchar realname[512];
+
+				gnksa_do_check_from (name,
+				                     address, sizeof(address),
+				                     realname, sizeof(realname),
+				                     FALSE);
 
-				gnksa_do_check_from (name, &address, &realname, FALSE);
 				okay = !smtp_send_line_va (item, sock, fmt, address);
-
-				g_free (address);
-				g_free (realname);
 			}
 			if (okay)
 			{
@@ -301,9 +307,9 @@ smtp_send_article (StatusItem    * item,
 	*** send the message
 	**/
         
-	references = article_get_header (article, HEADER_REFERENCES);
+	references = article->references;
 
-	reply_to = article_get_header (article, HEADER_REPLY_TO);
+	reply_to = article_get_extra_header (article, HEADER_REPLY_TO);
 
 	subject = article_get_subject (article);
 
@@ -331,6 +337,13 @@ smtp_send_article (StatusItem    * item,
 		okay = !smtp_send_line_va (item, sock, "Subject: %s\r\n", tmp);
 		g_free (tmp);
 	}
+	if (okay) {
+		gchar * date = rfc822_date_generate(0);
+		gchar * tmp = g_mime_utils_8bit_header_encode (date);
+		okay = !smtp_send_line_va (item, sock, "Date: %s\r\n", tmp);
+		g_free (tmp);
+		g_free (date);
+	}
 	if (okay && references!=NULL) {
 		gchar * tmp = g_mime_utils_8bit_header_encode (references);
 		okay = !smtp_send_line_va (item, sock, "References: %s\r\n", tmp);
@@ -472,7 +485,7 @@ smtp_send (StatusItem      * item,
 
 	/* sanity checks */
 	g_return_val_if_fail (item!=NULL, -1);
-	g_return_val_if_fail (article!=NULL, -1);
+	g_return_val_if_fail (article_is_valid(article), -1);
 
 	/* connect to the smtp server */
 	sock = smtp_get_connection (item);
@@ -487,3 +500,6 @@ smtp_send (StatusItem      * item,
 	pan_object_unref (PAN_OBJECT(sock));
 	return retval;
 }
+
+
+
