@@ -61,13 +61,15 @@ static void status_item_view_init (StatusItemView *object);
 static void status_item_view_destroy (GtkObject *object);
 static void status_item_view_shutdown (GtkObject *object);
 
-static gchar* string_shorten_nolock (GdkFont* font, int width, const gchar* original);
+static gchar* string_shorten_nolock (GdkFont     * font,
+                                     gint          width,
+                                     const gchar * original,
+                                     gint          original_len);
 
 static void status_item_view_unregister (StatusItemView *view);
 
 static void progress_cb      (gpointer, gpointer, gpointer);
 static void status_cb        (gpointer, gpointer, gpointer);
-static void status_cb_nolock (gpointer, gpointer, gpointer);
 
 /*********************
 **********************  Variables
@@ -97,52 +99,29 @@ static GtkObjectClass *parent_class = NULL;
 *************  PRIVATE ROUTINES
 ************/
 
-/**
- * Structure holding the arguments to gtk_proress_set_percentage.
- */
-typedef struct
-{
-	GtkProgress * progress;
-	gfloat f;
-}
-IdleStruct;
-
-/**
- * Call gtk_progress_set_percentage in an idle func.
- */
 static gint
-set_percentage_mainthread (gpointer p)
+set_percentage_mainthread (gpointer user_data)
 {
-	IdleStruct * data = (IdleStruct*) p;
+	ArgSet * argset = (ArgSet*) user_data;
+	StatusItemView * siv = (StatusItemView *)(argset_get (argset, 0));
+	const int of_100 = GPOINTER_TO_INT (argset_get (argset, 1));
 
-	pan_lock_unconditional ();
-	if (GTK_IS_PROGRESS(data->progress))
-	{
-		gtk_progress_set_percentage (data->progress, data->f);
-		gtk_widget_unref (GTK_WIDGET(data->progress));
-	}
-	pan_unlock_unconditional ();
+	/* update the percentage */
+	pan_lock ();
+	gtk_progress_set_percentage (GTK_PROGRESS(siv->progress), (gfloat)(of_100/100.0));
+	pan_unlock ();
 
-	g_free (data);
+	/* cleanup */
+	argset_free (argset);
 	return 0;
 }
 
-/**
- * Defer the call to an idle func to ensure we're in the right thread.
- */
 static void
-set_percentage_nolock (GtkProgress * progress, gfloat f)
+progress_cb (gpointer call_object, gpointer call_arg, gpointer user_data)
 {
-	IdleStruct * s;
-
-	g_return_if_fail (GTK_IS_PROGRESS(progress));
-
-	s = g_new0 (IdleStruct, 1);
-	s->progress = progress;
-	s->f = f;
-
-	gtk_widget_ref (GTK_WIDGET(progress));
-	run_in_main_thread_nolock (set_percentage_mainthread, s);
+	const gint of_100 = GPOINTER_TO_INT(call_arg);
+	if (0<=of_100 && of_100<=100)
+		gui_queue_add (set_percentage_mainthread, argset_new2 (user_data, GINT_TO_POINTER(of_100)));
 }
 
 /************
@@ -194,9 +173,8 @@ static void
 status_item_view_init (StatusItemView *view)
 {
 	view->progress = NULL;
-	view->last_status = NULL;
 	view->item = NULL;
-	view->of_100 = -1;
+	view->last_status = g_string_new (NULL);
 }
 
 
@@ -223,7 +201,8 @@ status_item_view_destroy (GtkObject* object)
 	g_return_if_fail (IS_STATUS_ITEM_VIEW (object));
 	g_return_if_fail (GTK_OBJECT_CONSTRUCTED (object));
 
-	replace_gstr (&item->last_status, NULL);
+	g_string_truncate (item->last_status, 0);
+	g_string_free (item->last_status, TRUE);
 
 	(*GTK_OBJECT_CLASS (parent_class)->destroy) (object);
 }
@@ -292,8 +271,8 @@ status_item_view_set_item_nolock (StatusItemView* view, StatusItem* item)
 		status_item_view_unregister (view);
 
 		/* update last_status; update progressbar w/o a resize */
-		set_percentage_nolock (prog, 0);
-		replace_gstr (&view->last_status, g_strdup(""));
+		gui_queue_add (set_percentage_mainthread, argset_new2 (view, NULL));
+		g_string_assign (view->last_status, "");
 		replace_gstr (&prog->format, g_strdup(""));
 		GTK_PROGRESS_CLASS (GTK_OBJECT(prog)->klass)->update (prog);
 		gtk_tooltips_set_tip (ttips, GTK_WIDGET(view->eventbox), NULL, NULL);
@@ -314,10 +293,9 @@ status_item_view_set_item_nolock (StatusItemView* view, StatusItem* item)
 		/* show the last status text */
 		if (item !=NULL)
 		{
-			gchar * pch = g_strdup (item->status_text);
+			gchar * pch = item->status_text;
 			if (is_nonempty_string(pch))
-				status_cb_nolock (NULL, pch, view);
-			g_free (pch);
+				status_cb (NULL, pch, view);
 		}
 	}
 
@@ -328,54 +306,36 @@ status_item_view_set_item_nolock (StatusItemView* view, StatusItem* item)
 ******
 *****/
 
-static void
-progress_cb (gpointer call_object, gpointer call_arg, gpointer user_data)
-{
-	const gint of_100 = GPOINTER_TO_INT(call_arg);
-	if (0<=of_100 && of_100<=100)
-	{
-		StatusItemView * siv = STATUS_ITEM_VIEW(user_data);
-		pan_lock ();
-		set_percentage_nolock (GTK_PROGRESS(siv->progress), of_100/100.0);
-		pan_unlock ();
-	}
-}
 
 static gint
-status_idle (gpointer data)
+status_cb_mainthread (gpointer data)
 {
 	StatusItemView * view;
 
-	pan_lock_unconditional ();
+	pan_lock ();
 	view = STATUS_ITEM_VIEW(data);
 	if (view!=NULL && view->item!=NULL)
 	{
 		GtkProgress * prog = GTK_PROGRESS(view->progress);
 		gchar * s = string_shorten_nolock (gtk_widget_get_style (view->progress)->font,
 		                                   view->progress->allocation.width,
-		                                   view->last_status);
+		                                   view->last_status->str,
+		                                   view->last_status->len);
 		replace_gstr (&prog->format, s);
 		/* update progressbar w/o a resize */
 		GTK_PROGRESS_CLASS (GTK_OBJECT(prog)->klass)->update (prog);
 	}
-	pan_unlock_unconditional ();
+	pan_unlock ();
 
 	return 0;
 }
 
-static void
-status_cb_nolock (gpointer call_object, gpointer call_arg, gpointer user_data)
-{
-	StatusItemView * siv = STATUS_ITEM_VIEW (user_data);
-	replace_gstr (&siv->last_status, g_strdup ((const gchar*)call_arg));
-	run_in_main_thread_nolock (status_idle, siv);
-}
 static void
 status_cb (gpointer call_object, gpointer call_arg, gpointer user_data)
 {
-	pan_lock ();
-	status_cb_nolock (call_object, call_arg, user_data);
-	pan_unlock ();
+	StatusItemView * siv = STATUS_ITEM_VIEW (user_data);
+	g_string_assign (siv->last_status, (const gchar*) call_arg);
+	gui_queue_add (status_cb_mainthread, siv);
 }
 
 /**
@@ -383,10 +343,9 @@ status_cb (gpointer call_object, gpointer call_arg, gpointer user_data)
 **/
 
 static void
-string_zotz_char (gchar   * str,
-                  gint      i)
+string_remove_char (gchar * str, gint index)
 {
-	gchar *p = str+i;
+	gchar *p = str+index;
 	for (;;) {
 		++p;
 		p[-1] = *p;
@@ -398,12 +357,14 @@ string_zotz_char (gchar   * str,
 static gchar*
 string_shorten_nolock (GdkFont           * font,
                        gint                max_width,
-                       const gchar       * original)
+                       const gchar       * original,
+                       gint                original_len)
 {
 	gchar *clone = NULL;
-	size_t len = 0;
+	size_t len;
 	gint i, j;
 	gboolean i_turn = TRUE;
+	const int DOTLEN = 3;
 	gint target_width = max_width - 4 - 30; /* 30 is just a magic safety number */
 
 	/* if original is NULL, so is shortened */
@@ -411,28 +372,30 @@ string_shorten_nolock (GdkFont           * font,
 		return NULL;
 
 	/* if original fits, use it */
-	len = strlen(original);
-	if (gdk_text_width (font, original, len) < target_width)
+	if (gdk_text_width (font, original, original_len) < target_width)
 		return g_strdup (original);
 
 	/* add in the middle "..." */
-	clone = g_malloc (len + strlen("...") + 1);
-	i = len/2;
+	len = original_len;
+	clone = g_malloc (original_len + DOTLEN + 1);
+	i = original_len/2;
 	memcpy (clone, original, i);
-	memcpy (clone+i, "...", 3);
+	memcpy (clone+i, "...", DOTLEN);
 	j = i+3;
-	memcpy (clone+j, original+i, len+1-i);
+	memcpy (clone+j, original+i, original_len+1-i);
 	original = NULL;
 
 	/* start trimming out text until it fits */	
-	while (gdk_text_width(font, clone, strlen(clone)) >= target_width)
+	len = original_len + DOTLEN;
+	while (gdk_text_width(font, clone, len) >= target_width)
 	{
 		int *pi = i_turn ? &i : &j;
 		*pi -= 1;
-		if (*pi<0 || *pi>=strlen(clone))
+		if (*pi<0 || *pi>=len)
 			break;
-		string_zotz_char (clone, *pi);
+		string_remove_char (clone, *pi);
 		i_turn = !i_turn;
+		--len;
 	}
 
 	return clone;
