@@ -188,6 +188,7 @@ task_body_run (Task * task)
 	const gboolean * abort;
 	const gchar * message_id;
 	gboolean is_folder;
+	GString * text;
 	debug_enter ("task_body_run");
 
 	/* initialize variables */
@@ -198,6 +199,7 @@ task_body_run (Task * task)
 	article = body->article;
 	abort = &task->hint_abort;
 	message_id = article_get_message_id (article);
+	text = NULL;
 	
 	/* sanity clause */
 	g_return_val_if_fail (server!=NULL, TASK_FAIL_HOPELESS);
@@ -238,13 +240,15 @@ task_body_run (Task * task)
 			return TASK_FAIL;
 		}
 
-		result = nntp_article_download (status,
-		                                task->sock,
-		                                group,
-		                                article,
-		                                &task->hint_abort,
-		                                NNTP_VERBOSE,
-		                                TRUE);
+		text = g_string_sized_new (400 * 80); /* prealloc space for ~400 lines */
+		result = nntp_article_download_to_gstr (status,
+		                                        task->sock,
+		                                        group,
+		                                        article,
+		                                        &task->hint_abort,
+		                                        NNTP_VERBOSE,
+		                                        TRUE,
+		                                        text);
 		if (result != TASK_SUCCESS)
 		{
 			status_item_emit_status_va (status,
@@ -266,14 +270,25 @@ task_body_run (Task * task)
 				g_string_free (tmp, TRUE);
 			}
 
+			g_string_free (text, TRUE);
 			debug_exit ("task_body_run");
 			return result;
 		}
 	}
 
-	/* read the message */
-	text_set_from_cached_article (article);
-	articles_set_read (&article, 1, TRUE);
+	/**
+	***  Update the text widget.
+	***  If we just downloaded it then we still have it in memory
+	***  so call _from_mem_article.  Otherwise call _from_cached_article.
+	**/
+
+	if (text != NULL) {
+		text_set_from_mem_article (article, text->str);
+		g_string_free (text, TRUE);
+	}
+	else if (article_has_body (article)) {
+		text_set_from_cached_article (article);
+	}
 
 	debug_exit ("task_body_run");
 	return TASK_SUCCESS;
