@@ -276,7 +276,6 @@ nntp_grouplist_download_all (TaskGroupList * item)
 			/* create a new group object */
 			group = group_new (server, name);
 			group->permission = *permission;
-			group_set_dirty (group); /* make sure it gets saved */
 			g_ptr_array_add (groups, group);
 
 			/* periodic status feedback */
@@ -294,6 +293,10 @@ nntp_grouplist_download_all (TaskGroupList * item)
 			g_free (permission);
 		}
 
+		/* make sure the new groups get saved */
+		if (groups->len)
+			server_set_group_type_dirty (server, SERVER_GROUPS_ALL);
+
 		if (!read_err)
 			item->got_list = TRUE;
 
