@@ -188,6 +188,36 @@ main (void)
 		g_message ("PASS: filenamename collision avoided");
 	}
 
+	/**
+	***  UUEncoded,  in Filename
+	**/
+
+	if (1)
+	{
+		GSList * sl = NULL;
+		gchar * text = "<micro>";
+		Article * article = article_new (group);
+		article_set_header (article, HEADER_SUBJECT, text, DO_CHUNK);
+		article_set_header (article, HEADER_MESSAGE_ID, text, DO_CHUNK);
+		article->part = 1;
+		article->parts = 1;
+		article->number = 1;
+		sl = g_slist_append (sl, article);
+
+		dd.server = server;
+		dd.group = group;
+		dd.articles = sl;
+		dd.path = download_dir;
+		dd.filename = NULL;
+		dd.item = status;
+		decode_article (&dd);
+		g_slist_free (sl);
+		sl = NULL;
+
+		check_md5 ("micro.txt", "39c14a46362de0f4764faef78e1a77b7");
+		g_message ("PASS: filename with ");
+	}
+
 	/**
 	***  UUEncoded, in Apostrophe Filename
 	**/
